"""
Viewer object cache implementation

Important to have because if we're debugging potential state management issues
in the viewer's scene graph, we need an idea of what it's scene graph _should_
look like at the current point in time. We can get that by hooking into its
VOCache so we know about its cache hits, and then compare whats in the proxy's
ObjectManager vs the viewer's (through GDB or something.)

Everything little-endian unless otherwise specified.
These use native struct alignment and padding, which is the reason for the
native address size being stored in the header. They should have just packed
the structs properly instead.

object.cache index file:
    IndexMetaHeader:
        U32 version = 15;
        U32 address_size = 32 or 64;
        CacheIndex entries[128];

    Exactly 128 region entries allowed, if any are missing they will have `time == 0`
    and should be skipped.

    CacheIndex:
        S32 index = i; // redundant, but helpful
        U64 handle; // ORed together global X and Y
        U32 time;


objects_<grid_x>_<grix_y>.slc:
    ObjectsMetaHeader:
        // must match ID sent in RegionHandshake. Filenames do not include grid ID so this may be
        // a file for a region at the same coords on a completely different grid!
        UUID cache_id;
        S32 num_entries;

    VOCacheEntry:
        U32 local_id;
        U32 crc;
        S32 hit_count;
        S32 dupe_count;
        S32 crc_change_count;
        // must be <= 10000 and > 0. Failing this continues parsing without reading data.
        S32 size;
        if (size <= 10000 && entry.size > 0)
            U8 data[size]; // same representation as "data" in ObjectUpdateCompressed
        else
            U8 data[0];


    ObjectsMetaHeader header;
    for i in range(header.num_entries) {
        VOCacheEntry entry;
    }
"""

from __future__ import annotations

import io
import logging
import pathlib
from pathlib import Path
from typing import *

import recordclass

import hippolyzer.lib.base.serialization as se
from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.objects import handle_to_gridxy
from hippolyzer.lib.proxy.viewer_settings import iter_viewer_cache_dirs

LOG = logging.getLogger(__name__)


class ViewerObjectCache:
    VERSION = 15
    MAX_REGIONS = 128

    def __init__(self, base_path: Union[str, Path]):
        self.base_path = Path(base_path)
        # handle -> updated
        self.regions: Dict[int, int] = {}

    @classmethod
    def from_path(cls, base_path: Union[str, Path]):
        base_path = pathlib.Path(base_path)
        cache = cls(base_path)
        with open(cache.base_path / "object.cache", "rb") as fh:
            reader = se.BufferReader("<", fh.read())
        version = reader.read(se.U32)
        if version != cls.VERSION:
            LOG.error(f"Unsupported vocache version {version} in {cache.base_path}")
            return
        address_size = reader.read(se.U32)
        if address_size not in (32, 64):
            LOG.error(f"Unsupported address size {address_size}")
            return

        # HACK: VOCache writes structs directly to disk from memory. It doesn't specify
        # any packing rules, so the struct gets written with whatever the platform
        # defaults are. In my case, everything is 8 byte aligned because there's a
        # U64 member for the handle. I'm not an expert in this sort of thing, so we
        # try to guess the arrangement of the struct by scanning ahead.
        int_spec = se.U32
        for i in range(cls.MAX_REGIONS):
            entry_index = reader.read(int_spec) & 0xFFffFFff
            if entry_index != i:
                LOG.warning(f"Expected region entry index to be {i}, got {entry_index}")
            # Sniff padding alignment on the first cache entry
            if i == 0:
                # Seek to where the next index would be if everything was 8 byte aligned
                with reader.scoped_seek(20, io.SEEK_CUR):
                    next_i = reader.read(se.U32)

                # If it's 1 then we're using 8 byte alignment. Just read 8 bytes for all ints.
                # If there was no padding then this would read into the region handle, but
                # that could never have 4 bytes == 1 because both x and y will be multiples of 256.
                if next_i == 1:
                    # Trash the extra few bits and switch to reading U64s
                    _ = reader.read(se.U32)
                    int_spec = se.U64

            handle = reader.read(se.U64)
            # Mask off any junk bits that might have been written in the padding
            time = reader.read(int_spec) & 0xFFffFFff
            # If there's no time then this is an empty slot.
            if not time:
                continue
            cache.regions[handle] = time
        return cache

    def read_region(self, handle: int) -> Optional[RegionViewerObjectCache]:
        if handle not in self.regions:
            return None
        grid_x, grid_y = handle_to_gridxy(handle)
        objects_file = self.base_path / f"objects_{grid_x}_{grid_y}.slc"
        if not objects_file.exists():
            return None
        return RegionViewerObjectCache.from_file(objects_file)


class ViewerObjectCacheEntry(recordclass.dataobject):   # type: ignore
    local_id: int
    crc: int
    data: bytes


def is_valid_vocache_dir(cache_dir):
    return (pathlib.Path(cache_dir) / "objectcache" / "object.cache").exists()


class RegionViewerObjectCache:
    """Parser and container for .slc files"""
    def __init__(self, cache_id: UUID, entries: List[ViewerObjectCacheEntry]):
        self.cache_id: UUID = cache_id
        self.entries: Dict[int, ViewerObjectCacheEntry] = {
            e.local_id: e for e in entries
        }

    @classmethod
    def from_file(cls, objects_path: Union[str, Path]):
        # These files are only a few megabytes max so fine to slurp in
        with open(objects_path, "rb") as fh:
            reader = se.BufferReader("<", fh.read())
        cache_id: UUID = reader.read(se.UUID)

        num_entries = reader.read(se.S32)
        entries = []
        for _ in range(num_entries):
            # EOF, the viewer specifically allows for this.
            if not len(reader):
                break
            local_id = reader.read(se.U32)
            crc = reader.read(se.U32)
            # Not important to us
            _ = reader.read(se.U32)
            _ = reader.read(se.U32)
            _ = reader.read(se.U32)
            size = reader.read(se.U32)
            if not size or size > 10_000:
                continue
            data = reader.read_bytes(size, to_bytes=True)
            entries.append(ViewerObjectCacheEntry(
                local_id=local_id,
                crc=crc,
                data=data,
            ))
        return RegionViewerObjectCache(cache_id, entries)

    def lookup_object_data(self, local_id: int, crc: int) -> Optional[bytes]:
        entry = self.entries.get(local_id)
        if entry and entry.crc == crc:
            return entry.data
        return None


class RegionViewerObjectCacheChain:
    """Wrapper for the checking the same region in multiple cache locations"""
    def __init__(self, region_caches: List[RegionViewerObjectCache]):
        self.region_caches = region_caches

    def lookup_object_data(self, local_id: int, crc: int) -> Optional[bytes]:
        for cache in self.region_caches:
            data = cache.lookup_object_data(local_id, crc)
            if data:
                return data
        return None

    @classmethod
    def for_region(cls, handle: int, cache_id: UUID, cache_dir: Optional[str] = None):
        """
        Get a cache chain for a specific region, called on region connection

        We don't know what viewer the user is currently using, or where its cache lives
        so we have to try every region object cache file for every viewer installed.
        """
        caches = []
        if cache_dir is None:
            cache_dirs = iter_viewer_cache_dirs()
        else:
            cache_dirs = [pathlib.Path(cache_dir)]

        for cache_dir in cache_dirs:
            if not is_valid_vocache_dir(cache_dir):
                continue
            cache = ViewerObjectCache.from_path(cache_dir / "objectcache")
            if cache:
                caches.append(cache)
        regions = []
        for cache in caches:
            region = cache.read_region(handle)
            if not region:
                continue
            if region.cache_id != cache_id:
                continue
            regions.append(region)
        return RegionViewerObjectCacheChain(regions)



from __future__ import annotations

import gzip
import logging
import secrets
from pathlib import Path
from typing import Union, List, Tuple, Set

from hippolyzer.lib.base import llsd
from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.inventory import InventoryModel, InventoryCategory, InventoryItem
from hippolyzer.lib.base.message.message import Block
from hippolyzer.lib.client.state import BaseClientSession


LOG = logging.getLogger(__name__)


class InventoryManager:
    def __init__(self, session: BaseClientSession):
        self._session = session
        self.model: InventoryModel = InventoryModel()
        self._load_skeleton()

    def _load_skeleton(self):
        assert not self.model.nodes
        skel_cats: List[dict] = self._session.login_data.get('inventory-skeleton', [])
        for skel_cat in skel_cats:
            self.model.add(InventoryCategory(
                name=skel_cat["name"],
                cat_id=UUID(skel_cat["folder_id"]),
                parent_id=UUID(skel_cat["parent_id"]),
                # Don't use the version from the skeleton, this flags the inventory as needing
                # completion from the inventory cache. This matches indra's behavior.
                version=InventoryCategory.VERSION_NONE,
                type="category",
                pref_type=skel_cat.get("type_default", -1),
                owner_id=self._session.agent_id,
            ))

    def load_cache(self, path: Union[str, Path]):
        # Per indra, rough flow for loading inv on login is:
        # 1. Look at inventory skeleton from login response
        # 2. Pre-populate model with categories from the skeleton, including their versions
        # 3. Read the inventory cache, tracking categories and items separately
        # 4. Walk the list of categories in our cache. If the cat exists in the skeleton and the versions
        #    match, then we may load the category and its descendants from cache.
        # 5. Any categories in the skeleton but not in the cache, or those with mismatched versions must be fetched.
        #    The viewer does this by setting the local version of the cats to -1 and forcing a descendent fetch
        #    over AIS.
        #
        # By the time you call this function call, you should have already loaded the inventory skeleton
        # into the model set its inventory category versions to VERSION_NONE.

        skel_cats: List[dict] = self._session.login_data['inventory-skeleton']
        # UUID -> version map for inventory skeleton
        skel_versions = {UUID(cat["folder_id"]): cat["version"] for cat in skel_cats}
        LOG.info(f"Parsing inv cache at {path}")
        cached_categories, cached_items = self._parse_cache(path)
        LOG.info(f"Done parsing inv cache at {path}")
        loaded_cat_ids: Set[UUID] = set()

        for cached_cat in cached_categories:
            existing_cat: InventoryCategory = self.model.get(cached_cat.cat_id)  # noqa
            # Don't clobber an existing cat unless it just has a placeholder version,
            # maybe from loading the skeleton?
            if existing_cat and existing_cat.version != InventoryCategory.VERSION_NONE:
                continue
            # Cached cat isn't the same as what the inv server says it should be, can't use it.
            if cached_cat.version != skel_versions.get(cached_cat.cat_id):
                continue
            if existing_cat:
                # Remove the category so that we can replace it, but leave any children in place
                self.model.unlink(existing_cat, single_only=True)
            self.model.add(cached_cat)
            # Any items in this category in our cache file are usable and should be added
            loaded_cat_ids.add(cached_cat.cat_id)

        for cached_item in cached_items:
            # The skeleton doesn't have any items, so if we run into any items they should be exactly the
            # same as what we're trying to add. No point clobbering.
            if cached_item.item_id in self.model:
                continue
            # The parent category didn't have a cache hit against the inventory skeleton, can't add!
            if cached_item.parent_id not in loaded_cat_ids:
                continue
            self.model.add(cached_item)

    def _parse_cache(self, path: Union[str, Path]) -> Tuple[List[InventoryCategory], List[InventoryItem]]:
        categories: List[InventoryCategory] = []
        items: List[InventoryItem] = []
        # Parse our cached items and categories out of the compressed inventory cache
        first_line = True
        with gzip.open(path, "rb") as f:
            # Line-delimited LLSD notation!
            for line in f.readlines():
                # TODO: Parsing of invcache is dominated by `parse_notation()`. It's stupidly inefficient.
                node_llsd = llsd.parse_notation(line)
                if first_line:
                    # First line is the file header
                    first_line = False
                    if node_llsd['inv_cache_version'] != 2:
                        raise ValueError(f"Unknown cache version: {node_llsd!r}")
                    continue

                if InventoryCategory.ID_ATTR in node_llsd:
                    if (cat_node := InventoryCategory.from_llsd(node_llsd)) is not None:
                        categories.append(cat_node)
                elif InventoryItem.ID_ATTR in node_llsd:
                    if (item_node := InventoryItem.from_llsd(node_llsd)) is not None:
                        items.append(item_node)
                else:
                    LOG.warning(f"Unknown node type in inv cache: {node_llsd!r}")
        return categories, items


# Thankfully we have 9 billion different ways to represent inventory data.
def ais_item_to_inventory_data(ais_item: dict) -> Block:
    return Block(
        "InventoryData",
        ItemID=ais_item["item_id"],
        FolderID=ais_item["parent_id"],
        CallbackID=0,
        CreatorID=ais_item["permissions"]["creator_id"],
        OwnerID=ais_item["permissions"]["owner_id"],
        GroupID=ais_item["permissions"]["group_id"],
        BaseMask=ais_item["permissions"]["base_mask"],
        OwnerMask=ais_item["permissions"]["owner_mask"],
        GroupMask=ais_item["permissions"]["group_mask"],
        EveryoneMask=ais_item["permissions"]["everyone_mask"],
        NextOwnerMask=ais_item["permissions"]["next_owner_mask"],
        GroupOwned=0,
        AssetID=ais_item["asset_id"],
        Type=ais_item["type"],
        InvType=ais_item["inv_type"],
        Flags=ais_item["flags"],
        SaleType=ais_item["sale_info"]["sale_type"],
        SalePrice=ais_item["sale_info"]["sale_price"],
        Name=ais_item["name"],
        Description=ais_item["desc"],
        CreationDate=ais_item["created_at"],
        # Meaningless here
        CRC=secrets.randbits(32),
    )


def inventory_data_to_ais_item(inventory_data: Block) -> dict:
    return dict(
        item_id=inventory_data["ItemID"],
        parent_id=inventory_data["ParentID"],
        permissions=dict(
            creator_id=inventory_data["CreatorID"],
            owner_id=inventory_data["OwnerID"],
            group_id=inventory_data["GroupID"],
            base_mask=inventory_data["BaseMask"],
            owner_mask=inventory_data["OwnerMask"],
            group_mask=inventory_data["GroupMask"],
            everyone_mask=inventory_data["EveryoneMask"],
            next_owner_mask=inventory_data["NextOwnerMask"],
        ),
        asset_id=inventory_data["AssetID"],
        type=inventory_data["Type"],
        inv_type=inventory_data["InvType"],
        flags=inventory_data["Flags"],
        sale_info=dict(
            sale_type=inventory_data["SaleType"],
            sale_price=inventory_data["SalePrice"],
        ),
        name=inventory_data["Name"],
        description=inventory_data["Description"],
        creation_at=inventory_data["CreationDate"],
    )


def ais_folder_to_inventory_data(ais_folder: dict) -> Block:
    return Block(
        "FolderData",
        FolderID=ais_folder["cat_id"],
        ParentID=ais_folder["parent_id"],
        CallbackID=0,
        Type=ais_folder["preferred_type"],
        Name=ais_folder["name"],
    )


def inventory_data_to_ais_folder(inventory_data: Block) -> dict:
    return dict(
        cat_id=inventory_data["FolderID"],
        parent_id=inventory_data["ParentID"],
        preferred_type=inventory_data["Type"],
        name=inventory_data["Name"],
    )



import calendar
import datetime
import struct
import typing
import uuid
import zlib

from llsd import *
# So we can directly reference the original wrapper funcs where necessary
import llsd as base_llsd
from llsd.base import is_string, is_unicode

from hippolyzer.lib.base.datatypes import *


class HippoLLSDBaseFormatter(base_llsd.base.LLSDBaseFormatter):
    UUID: callable
    ARRAY: callable

    def __init__(self):
        super().__init__()
        self.type_map[UUID] = self.UUID
        self.type_map[Vector2] = self.TUPLECOORD
        self.type_map[Vector3] = self.TUPLECOORD
        self.type_map[Vector4] = self.TUPLECOORD
        self.type_map[Quaternion] = self.TUPLECOORD

    def TUPLECOORD(self, v: TupleCoord):
        return self.ARRAY(v.data())


class HippoLLSDXMLFormatter(base_llsd.serde_xml.LLSDXMLFormatter, HippoLLSDBaseFormatter):
    def __init__(self):
        super().__init__()


class HippoLLSDXMLPrettyFormatter(base_llsd.serde_xml.LLSDXMLPrettyFormatter, HippoLLSDBaseFormatter):
    def __init__(self):
        super().__init__()


def format_pretty_xml(val: typing.Any):
    return HippoLLSDXMLPrettyFormatter().format(val)


def format_xml(val: typing.Any):
    return HippoLLSDXMLFormatter().format(val)


class HippoLLSDNotationFormatter(base_llsd.serde_notation.LLSDNotationFormatter, HippoLLSDBaseFormatter):
    def __init__(self):
        super().__init__()

    def STRING(self, v):
        # llbase's notation LLSD encoder isn't suitable for generating line-delimited
        # LLSD because the string formatter leaves \n unencoded, unlike indra's llcommon.
        # Add our own escaping rule.
        return super().STRING(v).replace(b"\n", b"\\n")


def format_notation(val: typing.Any):
    return HippoLLSDNotationFormatter().format(val)


def format_binary(val: typing.Any, with_header=True):
    val = _format_binary_recurse(val)
    if with_header:
        return b'<?llsd/binary?>\n' + val
    return val


# This is copied almost wholesale from https://bitbucket.org/lindenlab/llbase/src/master/llbase/llsd.py
# With a few minor changes to make serialization round-trip correctly. It's evil.
def _format_binary_recurse(something) -> bytes:
    """Binary formatter workhorse."""
    def _format_list(list_something):
        array_builder = [b'[' + struct.pack('!i', len(list_something))]
        for item in list_something:
            array_builder.append(_format_binary_recurse(item))
        array_builder.append(b']')
        return b''.join(array_builder)

    if something is None:
        return b'!'
    elif isinstance(something, LLSD):
        return _format_binary_recurse(something.thing)
    elif isinstance(something, bool):
        if something:
            return b'1'
        else:
            return b'0'
    elif isinstance(something, int):
        try:
            return b'i' + struct.pack('!i', something)
        except (OverflowError, struct.error) as exc:
            raise LLSDSerializationError(str(exc), something)
    elif isinstance(something, float):
        try:
            return b'r' + struct.pack('!d', something)
        except SystemError as exc:
            raise LLSDSerializationError(str(exc), something)
    elif isinstance(something, uuid.UUID):
        return b'u' + something.bytes
    elif isinstance(something, binary):
        return b'b' + struct.pack('!i', len(something)) + something
    elif is_string(something):
        if is_unicode(something):
            something = something.encode("utf8")
        return b's' + struct.pack('!i', len(something)) + something
    elif isinstance(something, uri):
        return b'l' + struct.pack('!i', len(something)) + something.encode("utf8")
    elif isinstance(something, datetime.datetime):
        return b'd' + struct.pack('<d', something.timestamp())
    elif isinstance(something, datetime.date):
        seconds_since_epoch = calendar.timegm(something.timetuple())
        return b'd' + struct.pack('<d', seconds_since_epoch)
    elif isinstance(something, (list, tuple)):
        return _format_list(something)
    elif isinstance(something, dict):
        map_builder = [b'{' + struct.pack('!i', len(something))]
        for key, value in something.items():
            if isinstance(key, str):
                key = key.encode("utf8")
            map_builder.append(b'k' + struct.pack('!i', len(key)) + key)
            map_builder.append(_format_binary_recurse(value))
        map_builder.append(b'}')
        return b''.join(map_builder)
    else:
        try:
            return _format_list(list(something))
        except TypeError:
            raise LLSDSerializationError(
                "Cannot serialize unknown type: %s (%s)" %
                (type(something), something))


class HippoLLSDBinaryParser(base_llsd.serde_binary.LLSDBinaryParser):
    def __init__(self):
        super().__init__()
        self._dispatch[ord('u')] = lambda: UUID(bytes=self._getc(16))
        self._dispatch[ord('d')] = self._parse_date

    def _parse_date(self):
        seconds = struct.unpack("<d", self._getc(8))[0]
        try:
            return datetime.datetime.fromtimestamp(seconds, tz=datetime.timezone.utc)
        except OverflowError as exc:
            # A garbage seconds value can cause utcfromtimestamp() to raise
            # OverflowError: timestamp out of range for platform time_t
            self._error(exc, -8)

    def _parse_string(self):
        # LLSD's C++ API lets you stuff binary in a string field even though it's only
        # meant to be allowed in binary fields. Happens in SLM files. Handle that case.
        bytes_val = self._parse_string_raw()
        try:
            return bytes_val.decode('utf-8')
        except UnicodeDecodeError:
            pass
        return bytes_val


def parse_binary(data: bytes):
    if data.startswith(b'<?llsd/binary?>'):
        data = data.split(b'\n', 1)[1]
    return HippoLLSDBinaryParser().parse(data)


def parse_xml(data: bytes):
    return base_llsd.parse_xml(data)


def parse_notation(data: bytes):
    return base_llsd.parse_notation(data)


def zip_llsd(val: typing.Any):
    return zlib.compress(format_binary(val, with_header=False), level=zlib.Z_BEST_COMPRESSION)


def unzip_llsd(data: bytes):
    return parse_binary(zlib.decompress(data))


def parse(data: bytes):
    # You always have to content-type sniff because the provided
    # content-type is usually nonsense.
    try:
        data = data.lstrip()
        if data.startswith(b'<?llsd/binary?>'):
            return parse_binary(data)
        elif data.startswith(b'<'):
            return parse_xml(data)
        else:
            return parse_notation(data)
    except KeyError as e:
        raise base_llsd.LLSDParseError('LLSD could not be parsed: %s' % (e,))
    except TypeError as e:
        raise base_llsd.LLSDParseError('Input stream not of type bytes. %s' % (e,))



"""
Copyright 2009, Linden Research, Inc.
  See NOTICE.md for previous contributors
Copyright 2021, Salad Dais
All Rights Reserved.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 3 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with this program; if not, write to the Free Software Foundation,
Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
from __future__ import annotations

import copy
import enum
import importlib
import itertools
import logging
import os
import uuid
from typing import *

from hippolyzer.lib.base.datatypes import *
import hippolyzer.lib.base.serialization as se
import hippolyzer.lib.base.templates as templates
from hippolyzer.lib.base.datatypes import Pretty
from hippolyzer.lib.base.message.msgtypes import PacketFlags
from hippolyzer.lib.base.network.transport import Direction, ADDR_TUPLE

BLOCK_DICT = Dict[str, "MsgBlockList"]
VAR_TYPE = Union[TupleCoord, bytes, str, float, int, Tuple, UUID]

_TEMPLATES_MTIME = os.stat(templates.__file__).st_mtime


def maybe_reload_templates():
    # Templates may be modified at runtime during development, check
    # if they've changed since startup and reload if they have.
    global _TEMPLATES_MTIME
    templates_mtime = os.stat(templates.__file__).st_mtime

    if _TEMPLATES_MTIME is None or _TEMPLATES_MTIME < templates_mtime:
        print("Reloading templates")
        try:
            importlib.reload(templates)  # type: ignore
            _TEMPLATES_MTIME = templates_mtime
        except:
            logging.exception("Failed to reload templates!")


class Block:
    """
    base representation of a block
    Block expects a name, and kwargs for variables (var_name = value)
    """
    __slots__ = ('name', 'size', 'vars', 'message_name', '_ser_cache', 'fill_missing',)
    PARENT_MESSAGE_NAME: ClassVar[Optional[str]] = None

    def __init__(self, name, /, *, fill_missing=False, **kwargs):
        self.name = name
        self.size = 0
        self.message_name: Optional[str] = self.PARENT_MESSAGE_NAME
        self.vars: Dict[str, VAR_TYPE] = {}
        self._ser_cache: Dict[str, Any] = {}
        self.fill_missing = fill_missing
        for var_name, val in kwargs.items():
            self[var_name] = val

    def get_variable(self, var_name):
        return self.vars.get(var_name)

    def __contains__(self, item):
        return item in self.vars

    def __getitem__(self, name):
        return self.vars[name]

    def __setitem__(self, key, value):
        if isinstance(value, Pretty):
            return self.serialize_var(key, value.value)

        # These don't pickle well since they're likely to get hot-reloaded
        if isinstance(value, (enum.IntEnum, enum.IntFlag)):
            value = int(value)

        self.vars[key] = value
        # Invalidate the serialization cache if we manually changed the value
        if key in self._ser_cache:
            self._ser_cache.pop(key)

    def get_serializer(self, var_name) -> se.BaseSubfieldSerializer:
        serializer_key = (self.message_name, self.name, var_name)
        serializer = se.SUBFIELD_SERIALIZERS.get(serializer_key)
        if not serializer:
            raise KeyError(f"No serializer for {serializer_key!r}")
        return serializer

    def deserialize_var(self, var_name, make_copy=True):
        """
        Deserialize a var, using a cached version if available

        Does a deepcopy() of the value from the cache by default, so
        callers don't accidentally mutate the cached version. Allows
        opting out as otherwise deepcopy() can dominate runtime if
        you don't expect mutations.
        """
        if var_name in self._ser_cache:
            val = self._ser_cache[var_name]
            return copy.deepcopy(val) if make_copy else val
        serializer = self.get_serializer(var_name)
        deser = serializer.deserialize(self, self[var_name], pod=False)
        self._ser_cache[var_name] = deser
        return copy.deepcopy(deser) if make_copy else deser

    def serialize_var(self, var_name, val):
        serializer = self.get_serializer(var_name)
        serialized = serializer.serialize(self, val)
        self[var_name] = serialized
        self._ser_cache[var_name] = val

    def invalidate_caches(self):
        self._ser_cache.clear()

    def items(self):
        return self.vars.items()

    def finalize(self):
        # Stupid hack around the fact that blocks don't know how to
        # invoke field-specific serializers until they're added to a message.
        # Can do `Block("FooBlock", Baz_={"someserializedval": 1})` to set
        # "Baz" to a serialized val without having to first construct a message.
        for name in tuple(self.vars.keys()):
            if not name.endswith("_"):
                continue
            val = self.vars.pop(name)
            self.serialize_var(name.rstrip("_"), val)

    def repr(self, pretty=False):
        block_vars = {}
        if pretty:
            if not self.message_name:
                raise ValueError("Can't give pretty representation of block outside a message")
            for key in tuple(self.vars.keys()):
                try:
                    self.get_serializer(key)
                except KeyError:
                    block_vars[key] = repr(self.vars[key])
                    continue
                # We have a serializer, include the pretty output in the repr,
                # using the _ suffix so the builder knows it needs to be serialized.
                block_vars[f"{key}_"] = repr(self.deserialize_var(key))
        else:
            block_vars = self.vars

        kws = ", ".join((f"{k}={v if pretty else repr(v)}" for k, v in block_vars.items()))
        if kws:
            kws = ", " + kws
        return f"{self.__class__.__name__}({self.name!r}{kws})"

    def __repr__(self):
        return self.repr()


class MsgBlockList(List["Block"]):
    __slots__ = ()

    def __getitem__(self, item) -> Union["Block", VAR_TYPE]:
        if isinstance(item, str):
            return super().__getitem__(0)[item]
        return super().__getitem__(item)

    def __setitem__(self, item: Union[str, int, slice], val):
        if isinstance(item, str):
            self[0][item] = val
        else:
            super().__setitem__(item, val)


class Message:
    __slots__ = ("name", "send_flags", "packet_id", "acks", "body_boundaries", "queued",
                 "offset", "raw_extra", "raw_body", "deserializer", "_blocks", "finalized",
                 "direction", "meta", "synthetic", "dropped", "sender")

    def __init__(self, name, *args, packet_id=None, flags=0, acks=None, direction=None):
        # TODO: Do this on a timer or something.
        maybe_reload_templates()

        self.name = name
        self.send_flags = flags
        self.packet_id: Optional[int] = packet_id  # aka, sequence number

        self.acks = acks if acks is not None else tuple()
        self.body_boundaries = (-1, -1)
        self.offset = 0
        self.raw_extra = b""
        self.direction: Direction = direction if direction is not None else Direction.OUT
        # For lazy deserialization
        self.raw_body = None
        self.deserializer = None
        # should be set once a packet is sent / dropped to prevent accidental
        # re-sending or re-dropping
        self.finalized = False
        # Whether message is owned by the queue or should be sent immediately
        self.queued: bool = False
        self._blocks: BLOCK_DICT = {}
        self.meta = {}
        self.synthetic = packet_id is None
        self.dropped = False
        self.sender: Optional[ADDR_TUPLE] = None

        self.add_blocks(args)

    def add_blocks(self, block_list):
        # can have a list of blocks if it is multiple or variable
        for block in block_list:
            if type(block) == list:
                for bl in block:
                    self.add_block(bl)
            else:
                self.add_block(block)

    @property
    def extra(self) -> bytes:
        return self.raw_extra

    @extra.setter
    def extra(self, val: bytes):
        # Make sure this message is fully parsed if it wasn't already,
        # changing `.extra` requires re-serializing the message body.
        self.ensure_parsed()
        self.raw_extra = val
        self.offset = len(val)

    @property
    def blocks(self) -> BLOCK_DICT:
        self.ensure_parsed()
        return self._blocks

    @blocks.setter
    def blocks(self, val: BLOCK_DICT):
        self._blocks = val

        # block list was clobbered, so we don't care about any unparsed data
        self.raw_body = None
        self.deserializer = None

    def create_block_list(self, block_name: str):
        # There's a slight semantic difference between a missing block list
        # and a present block list with 0 length. This helps us distinguish.
        if block_name not in self.blocks:
            self.blocks[block_name] = MsgBlockList()

    def add_block(self, block: Block):
        self.create_block_list(block.name)

        self.blocks[block.name].append(block)
        block.message_name = self.name
        block.finalize()

    def get_block(self, block_name: str, default=None, /) -> Optional[Block]:
        return self.blocks.get(block_name, default)

    @property
    def reliable(self):
        # int() because otherwise this causes an alloc
        return self.send_flags & int(PacketFlags.RELIABLE)

    @property
    def has_acks(self):
        return self.send_flags & int(PacketFlags.ACK)

    @property
    def zerocoded(self):
        return self.send_flags & int(PacketFlags.ZEROCODED)

    @property
    def resent(self):
        return self.send_flags & int(PacketFlags.RESENT)

    def ensure_parsed(self):
        # This is a little magic, think about whether we want this.
        if self.raw_body and self.deserializer():
            self.deserializer().parse_message_body(self)

    def to_dict(self, extended=False):
        """ A dict representation of a message.

        This is the form used for templated messages sent via EQ.
        """
        self.ensure_parsed()
        base_repr = {'message': self.name, 'body': {}}

        for block_type in self.blocks:
            dict_blocks = base_repr['body'].setdefault(block_type, [])
            for block in self.blocks[block_type]:
                new_vars = {}
                for var_name, val in block.items():
                    new_vars[var_name] = val
                dict_blocks.append(new_vars)

        if extended:
            base_repr.update({
                "packet_id": self.packet_id,
                "meta": self.meta.copy(),
                "dropped": self.dropped,
                "synthetic": self.synthetic,
                "direction": self.direction.name,
                "send_flags": int(self.send_flags),
                "extra": self.extra,
                "acks": self.acks,
            })

        return base_repr

    @classmethod
    def from_dict(cls, dict_val):
        msg = cls(dict_val['message'])
        for block_type, blocks in dict_val['body'].items():
            msg.create_block_list(block_type)
            for block in blocks:
                msg.add_block(Block(block_type, **block))

        if 'packet_id' in dict_val:
            # extended format
            msg.packet_id = dict_val['packet_id']
            msg.meta = dict_val['meta']
            msg.dropped = dict_val['dropped']
            msg.synthetic = dict_val['synthetic']
            msg.direction = Direction[dict_val['direction']]
            msg.send_flags = dict_val['send_flags']
            msg.extra = dict_val['extra']
            msg.acks = dict_val['acks']
        return msg

    def invalidate_caches(self):
        # Don't have any caches if we haven't even parsed
        if self.raw_body:
            return
        for blocks in self.blocks.values():
            for block in blocks:
                block.invalidate_caches()

    def __getitem__(self, item):
        return self.blocks[item]

    def __setitem__(self, key: str, value):
        if not isinstance(value, (list, tuple)):
            value = (value,)
        if not isinstance(value, MsgBlockList):
            value = MsgBlockList(value)
        self.blocks[key] = value

    def __contains__(self, item):
        return item in self.blocks

    def _args_repr(self, pretty=False):
        sep = ",\n  "
        block_reprs = sep.join(x.repr(pretty=pretty) for x in itertools.chain(*self.blocks.values()))
        if block_reprs:
            block_reprs = sep + block_reprs
        return f"{self.name!r}{block_reprs}, direction=Direction.{self.direction.name}"

    def repr(self, pretty=False):
        self.ensure_parsed()
        return f"{self.__class__.__name__}({self._args_repr(pretty=pretty)})"

    def take(self):
        message_copy = copy.deepcopy(self)

        # Set the queued flag so the original will be dropped and acks will be sent
        if not self.finalized:
            self.queued = True

        # Original was dropped so let's make sure we have clean acks and packet id
        message_copy.acks = tuple()
        message_copy.send_flags &= ~PacketFlags.ACK
        message_copy.packet_id = None
        message_copy.dropped = False
        message_copy.finalized = False
        message_copy.queued = False
        return message_copy

    def to_summary(self):
        string = ""
        for block_name, block_list in self.blocks.items():
            for block in block_list:
                for var_name, val in block.items():
                    if block.name == "AgentData" and var_name in ("AgentID", "SessionID"):
                        continue
                    if string:
                        string += ", "
                    string += f"{var_name}={_trunc_repr(val, 10)}"
        return string

    def __repr__(self):
        return self.repr()

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.to_dict() == other.to_dict()


def _trunc_repr(val, max_len):
    if isinstance(val, (uuid.UUID, TupleCoord)):
        val = str(val)
    repr_val = repr(val)
    if isinstance(val, str):
        repr_val = repr_val[1:-1]
    if isinstance(val, bytes):
        repr_val = repr_val[2:-1]
    if len(repr_val) > max_len:
        return repr_val[:max_len] + "…"
    return repr_val



import datetime as dt

from hippolyzer.lib.base.helpers import get_mtime
from hippolyzer.lib.client.inventory_manager import InventoryManager
from hippolyzer.lib.client.state import BaseClientSession
from hippolyzer.lib.proxy.viewer_settings import iter_viewer_cache_dirs


class ProxyInventoryManager(InventoryManager):
    def __init__(self, session: BaseClientSession):
        super().__init__(session)
        newest_cache = None
        newest_timestamp = dt.datetime(year=1970, month=1, day=1, tzinfo=dt.timezone.utc)
        # Look for the newest version of the cached inventory and use that.
        # Not foolproof, but close enough if we're not sure what viewer is being used.
        for cache_dir in iter_viewer_cache_dirs():
            inv_cache_path = cache_dir / (str(session.agent_id) + ".inv.llsd.gz")
            if inv_cache_path.exists():
                mod = get_mtime(inv_cache_path)
                if not mod:
                    continue
                mod_ts = dt.datetime.fromtimestamp(mod, dt.timezone.utc)
                if mod_ts <= newest_timestamp:
                    continue
                newest_cache = inv_cache_path

        if newest_cache:
            self.load_cache(newest_cache)



import ctypes
import multiprocessing
import os
import sys
import time

SYNCHRONIZE = 0x100000


def _windows_pid_alive(pid: int):
    # https://stackoverflow.com/a/17645146
    # This is not quite right and https://stackoverflow.com/a/23409343 would
    # be better, but GetExitCodeProcess is always failing for me. Without
    # it we'll incorrectly treat a pid as alive for a few seconds after
    # it dies, but that's not a problem.
    kernel32 = ctypes.windll.kernel32
    process = kernel32.OpenProcess(SYNCHRONIZE, 0, pid)
    if process != 0:
        kernel32.CloseHandle(process)
        return True
    return False


class ParentProcessWatcher:
    def __init__(self, shutdown_signal: multiprocessing.Event):
        self._orig_ppid = os.getppid()
        self._win_last_pid_check = time.time()
        self._shutdown_signal = shutdown_signal

    def check_shutdown_needed(self) -> bool:
        if self._shutdown_signal.is_set():
            return True
        # Parent process died and we got reparented, set the shutdown signal
        # Even though we're a daemonized child process the parent doesn't get a chance
        # to kill us when it segfaults or gets a SIGKILL. prctl() is an option, but
        # *NIX-only. atexit() hooks also won't work here. Oh well.
        if os.getppid() != self._orig_ppid:
            self._shutdown_signal.set()
            return True
        # ppid won't change when parent dies on windows. have to do something worse.
        if sys.platform == "win32":
            # Only once per second, ctypes is slow.
            if self._win_last_pid_check + 1.0 < time.time():
                if not _windows_pid_alive(os.getppid()):
                    self._shutdown_signal.set()
                    return True
                self._win_last_pid_check = time.time()
        return False



import ast
import base64
import logging
import math
import re
import uuid
from typing import *

from .. import datatypes
from .. import llsd
from .. import serialization as se
from ..helpers import HippoPrettyPrinter
from ..network.transport import Direction
from .msgtypes import PacketFlags, MsgBlockType
from .template import MessageTemplate
from .message import Message, Block, maybe_reload_templates


class VerbatimHumanVal(str):
    pass


def _filtered_exports(mod):
    return {k: getattr(mod, k) for k in mod.__all__}


def subfield_eval(eval_str: str, globals_=None, locals_=None):
    return eval(
        eval_str,
        {
            "llsd": llsd,
            "base64": base64,
            "math": math,
            **_filtered_exports(datatypes),
            **(globals_ or {})},
        locals_
    )


TextSpan = Tuple[int, int]
SpanDict = Dict[Tuple[Union[str, int], ...], TextSpan]


class SpannedString(str):
    spans: SpanDict = {}


class HumanMessageSerializer:
    @classmethod
    def from_human_string(cls, string, replacements=None, env=None, safe=True):
        maybe_reload_templates()
        replacements = replacements or {}
        env = env or {}
        first_line = True
        cur_block = None
        msg = None
        lines = [x.strip() for x in string.split("\n") if x.strip()]
        while lines:
            line = lines.pop(0)
            # Ignore comment / blank lines
            if re.match(r"^\s*(#.*)?$", line):
                continue

            if first_line:
                first_split = [x for x in line.split(" ") if x]
                direction, message_name = first_split[:2]
                options = [x.strip("[]") for x in first_split[2:]]
                msg = Message(message_name)
                msg.direction = Direction[direction.upper()]
                for option in options:
                    if option in PacketFlags.__members__:
                        msg.send_flags |= PacketFlags[option]
                    elif re.match(r"^\d+$", option):
                        msg.send_flags |= int(option)
                first_line = False
                continue

            if line.startswith("["):
                cur_block = Block(re.search(r"\w+", line).group(0))
                msg.add_block(cur_block)
            else:
                expr_match = re.match(r"^\s*(\w+)\s*(=[|$]*)\s*(.*)$", line)
                var_name, operator, var_val = expr_match.groups()
                # Multiline, eat all the line continuations
                while var_val.endswith("\\"):
                    var_val = var_val[:-1].rstrip()
                    if lines:
                        var_val += lines.pop(0)

                plain = operator == "="
                packed = "|" in operator
                evaled = "$" in operator

                if evaled and safe:
                    raise ValueError("Can't use eval operator in safe mode")

                if plain:
                    replacement_match = re.match(r"\[\[(\w+)]]", var_val)
                    if replacement_match:
                        replacement_name = replacement_match.group(1)
                        var_val = replacements.get(replacement_name)
                        if var_val is None:
                            raise ValueError("Tried to use undefined replacement %s" % replacement_name)
                        if callable(var_val):
                            var_val = var_val()
                    # alternate way of specifying a vector or quat
                    elif var_val.startswith("<"):
                        var_val = re.sub(r"[<>]", "", var_val)
                        var_val = tuple(float(x) for x in var_val.split(","))
                    # UUID-ish
                    elif re.match(r"\A\w+-\w+-.*", var_val):
                        var_val = datatypes.UUID(var_val)
                    else:
                        var_val = ast.literal_eval(var_val)

                # Normally gross, but necessary for expressiveness in built messages
                # unless a metalanguage is added.
                if evaled:
                    var_val = subfield_eval(
                        var_val,
                        globals_={**env, **replacements},
                        locals_={"block": cur_block}
                    )
                # Using an packer specific to this message
                if packed:
                    if not evaled:
                        var_val = ast.literal_eval(var_val)
                    ser_key = (msg.name, cur_block.name, var_name)
                    serializer = se.SUBFIELD_SERIALIZERS.get(ser_key)
                    if not serializer:
                        raise KeyError(f"No subfield serializer for {ser_key!r}")
                    var_val = serializer.serialize(cur_block, var_val)

                cur_block[var_name] = var_val
        return msg

    @classmethod
    def to_human_string(cls, msg: Message, replacements=None, beautify=False,
                        template: Optional[MessageTemplate] = None) -> SpannedString:
        replacements = replacements or {}
        maybe_reload_templates()
        spans: SpanDict = {}
        string = ""
        if msg.direction is not None:
            string += f'{msg.direction.name} '
        string += msg.name
        flags = msg.send_flags
        for poss_flag in iter(PacketFlags):
            if flags & poss_flag:
                flags &= ~poss_flag
                string += f" [{poss_flag.name}]"
        # Make sure flags with unknown meanings don't get lost
        if flags:
            string += f" [{int(flags)}]"
        if msg.packet_id is not None:
            string += f'\n# ID: {msg.packet_id}'
            string += f'{", DROPPED" if msg.dropped else ""}{", SYNTHETIC" if msg.synthetic else ""}'
        if msg.extra:
            string += f'\n# EXTRA: {msg.extra!r}'
        string += '\n\n'

        for block_name, block_list in msg.blocks.items():
            block_suffix = ""
            if template and template.get_block(block_name).block_type == MsgBlockType.MBT_VARIABLE:
                block_suffix = '  # Variable'
            for block_num, block in enumerate(block_list):
                string += f"[{block_name}]{block_suffix}\n"
                for var_name, val in block.items():
                    start_len = len(string)
                    string += cls._format_var(msg, block, var_name, val, replacements, beautify)
                    end_len = len(string)
                    # Store the spans for each var so we can highlight specific matches
                    spans[(msg.name, block_name, block_num, var_name)] = (start_len, end_len)
                    string += "\n"
        spanned = SpannedString(string)
        spanned.spans = spans
        return spanned

    @classmethod
    def _format_var(cls, msg, block, var_name, var_val, replacements, beautify=False):
        string = ""
        # Check if we have a more human-readable way to present this field
        ser_key = (msg.name, block.name, var_name)
        serializer = se.SUBFIELD_SERIALIZERS.get(ser_key)
        field_prefix = ""
        if isinstance(var_val, VerbatimHumanVal):
            var_data = var_val
        elif isinstance(var_val, (uuid.UUID, datatypes.TupleCoord)):
            var_data = str(var_val)
        elif isinstance(var_val, (str, bytes)) and not serializer:
            var_data = cls._multi_line_pformat(var_val)
        else:
            var_data = repr(var_val)
        if serializer and beautify and not isinstance(var_val, VerbatimHumanVal):
            try:
                pretty_data = serializer.deserialize(block, var_val, pod=True)
                if pretty_data is not se.UNSERIALIZABLE:
                    string += f"  {var_name} =| {cls._multi_line_pformat(pretty_data)}"
                    if serializer.AS_HEX and isinstance(var_val, int):
                        var_data = hex(var_val)
                    if serializer.ORIG_INLINE:
                        string += f" #{var_data}"
                        return string
                    else:
                        string += "\n"
                    # Human-readable version should be used, orig data is commented out
                    field_prefix = "#"
            except:
                logging.exception(f"Failed in subfield serializer {ser_key!r}")
        if beautify:
            if block.name == "AgentData":
                if var_name == "AgentID" and var_val == replacements.get("AGENT_ID"):
                    var_data = "[[AGENT_ID]]"
                elif var_name == "SessionID" and var_val == replacements.get("SESSION_ID"):
                    var_data = "[[SESSION_ID]]"
            if "CircuitCode" in var_name or ("Code" in var_name and "Circuit" in block.name):
                if var_val == replacements.get("CIRCUIT_CODE"):
                    var_data = "[[CIRCUIT_CODE]]"
        string += f"  {field_prefix}{var_name} = {var_data}"
        return string

    @staticmethod
    def _multi_line_pformat(val):
        printer = HippoPrettyPrinter(width=100)
        val = printer.pformat(val)
        newstr = ""
        # Now we need to rebuild this to add in the appropriate
        # line continuations.
        lines = list(val.splitlines())
        first_line = True
        while lines:
            line = lines.pop(0)
            prefix = ""
            suffix = ""
            if first_line:
                first_line = False
            else:
                prefix = "    "

            if lines:
                suffix = " \\\n"
            newstr += f"{prefix}{line}{suffix}"
        return newstr



from __future__ import annotations

import dataclasses
import weakref
from typing import *

import transformations
from lxml import etree

from hippolyzer.lib.base.datatypes import Vector3, RAD_TO_DEG
from hippolyzer.lib.base.helpers import get_resource_filename


MAYBE_JOINT_REF = Optional[Callable[[], "JointNode"]]
SKELETON_REF = Optional[Callable[[], "Skeleton"]]


@dataclasses.dataclass
class JointNode:
    name: str
    parent: MAYBE_JOINT_REF
    skeleton: SKELETON_REF
    translation: Vector3
    pivot: Vector3  # pivot point for the joint, generally the same as translation
    rotation: Vector3  # Euler rotation in degrees
    scale: Vector3
    type: str  # bone or collision_volume

    def __hash__(self):
        return hash((self.name, self.type))

    @property
    def matrix(self):
        return transformations.compose_matrix(
            scale=tuple(self.scale),
            angles=tuple(self.rotation / RAD_TO_DEG),
            translate=tuple(self.translation),
        )

    @property
    def index(self) -> int:
        bone_idx = 0
        for node in self.skeleton().joint_dict.values():
            if node.type != "bone":
                continue
            if self is node:
                return bone_idx
            bone_idx += 1
        raise KeyError(f"{self.name!r} doesn't exist in skeleton")

    @property
    def ancestors(self) -> Sequence[JointNode]:
        joint_node = self
        ancestors = []
        while joint_node.parent:
            joint_node = joint_node.parent()
            ancestors.append(joint_node)
        return ancestors

    @property
    def children(self) -> Sequence[JointNode]:
        children = []
        for node in self.skeleton().joint_dict.values():
            if node.parent and node.parent() == self:
                children.append(node)
        return children

    @property
    def descendents(self) -> Set[JointNode]:
        descendents = set()
        ancestors = {self}
        last_ancestors = set()
        while last_ancestors != ancestors:
            last_ancestors = ancestors
            for node in self.skeleton().joint_dict.values():
                if node.parent and node.parent() in ancestors:
                    ancestors.add(node)
                    descendents.add(node)
        return descendents


class Skeleton:
    def __init__(self, root_node: etree.ElementBase):
        self.joint_dict: Dict[str, JointNode] = {}
        self._parse_node_children(root_node, None)

    def __getitem__(self, item: str) -> JointNode:
        return self.joint_dict[item]

    def _parse_node_children(self, node: etree.ElementBase, parent: MAYBE_JOINT_REF):
        name = node.get('name')
        joint = JointNode(
            name=name,
            parent=parent,
            skeleton=weakref.ref(self),
            translation=_get_vec_attr(node, "pos", Vector3()),
            pivot=_get_vec_attr(node, "pivot", Vector3()),
            rotation=_get_vec_attr(node, "rot", Vector3()),
            scale=_get_vec_attr(node, "scale", Vector3(1, 1, 1)),
            type=node.tag,
        )
        self.joint_dict[name] = joint
        for child in node.iterchildren():
            self._parse_node_children(child, weakref.ref(joint))


def _get_vec_attr(node, attr_name: str, default: Vector3) -> Vector3:
    attr_val = node.get(attr_name, None)
    if not attr_val:
        return default
    return Vector3(*(float(x) for x in attr_val.split(" ") if x))


def load_avatar_skeleton() -> Skeleton:
    skel_path = get_resource_filename("lib/base/data/avatar_skeleton.xml")
    with open(skel_path, 'r') as f:
        skel_root = etree.fromstring(f.read())
    return Skeleton(skel_root.getchildren()[0])


AVATAR_SKELETON = load_avatar_skeleton()



import dataclasses
from typing import *

from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.message.message import Message
from hippolyzer.lib.base.message.message_handler import MessageHandler


@dataclasses.dataclass
class NameCacheEntry:
    full_id: UUID
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    display_name: Optional[str] = None

    def __str__(self):
        if self.display_name:
            return f"{self.display_name} ({self.legacy_name})"
        if self.legacy_name:
            return self.legacy_name
        return f"(???) ({self.full_id})"

    @property
    def legacy_name(self) -> Optional[str]:
        if self.first_name is None:
            return None
        return f"{self.first_name} {self.last_name}"

    @property
    def preferred_name(self) -> Optional[str]:
        if self.display_name:
            return self.display_name
        return self.legacy_name


class NameCache:
    def __init__(self):
        self._cache: Dict[UUID, NameCacheEntry] = {}

    def create_subscriptions(
            self,
            message_handler: MessageHandler[Message, str],
    ):
        message_handler.subscribe("UUIDNameReply", self._handle_uuid_name_reply)

    def lookup(self, uuid: UUID, create_if_none: bool = False) -> Optional[NameCacheEntry]:
        val = self._cache.get(uuid)
        if create_if_none and val is None:
            val = NameCacheEntry(full_id=uuid)
            self._cache[uuid] = val
        return val

    def update(self, full_id: UUID, vals: dict):
        # upsert the cache entry
        entry = self._cache.get(full_id) or NameCacheEntry(full_id=full_id)
        if "FirstName" in vals:
            entry.first_name = vals["FirstName"]
        if "LastName" in vals:
            entry.last_name = vals["LastName"]
        if "DisplayName" in vals:
            entry.display_name = vals["DisplayName"] if vals["DisplayName"] else None
        self._cache[full_id] = entry

    def _handle_uuid_name_reply(self, msg: Message):
        for block in msg.blocks["UUIDNameBlock"]:
            self.update(block["ID"], {
                "FirstName": block["FirstName"],
                "LastName": block["LastName"],
            })

    def _process_display_names_response(self, parsed: dict):
        """Handle the response from the GetDisplayNames cap"""
        for agent in parsed["agents"]:
            # Don't set display name if they just have the default
            display_name = None
            if not agent["is_display_name_default"]:
                display_name = agent["display_name"]
            self.update(agent["id"], {
                "FirstName": agent["legacy_first_name"],
                "LastName": agent["legacy_last_name"],
                "DisplayName": display_name,
            })



import asyncio
import base64
import dataclasses
import email
import functools
import html
import itertools
import json
import logging
import pathlib
import multiprocessing
import re
import signal
import socket
import sys
import urllib.parse
from typing import *

import multidict
from qasync import QEventLoop, asyncSlot
from PySide6 import QtCore, QtWidgets, QtGui

from hippolyzer.apps.model import MessageLogModel, MessageLogHeader, RegionListModel
from hippolyzer.apps.proxy import start_proxy
from hippolyzer.lib.base import llsd
from hippolyzer.lib.base.datatypes import UUID
from hippolyzer.lib.base.helpers import bytes_unescape, bytes_escape, get_resource_filename
from hippolyzer.lib.base.message.llsd_msg_serializer import LLSDMessageSerializer
from hippolyzer.lib.base.message.message import Block, Message
from hippolyzer.lib.base.message.message_formatting import (
    HumanMessageSerializer,
    VerbatimHumanVal,
    subfield_eval,
    SpannedString,
)
from hippolyzer.lib.base.message.msgtypes import MsgType
from hippolyzer.lib.base.message.template_dict import DEFAULT_TEMPLATE_DICT
from hippolyzer.lib.base.settings import SettingDescriptor
from hippolyzer.lib.base.ui_helpers import loadUi
import hippolyzer.lib.base.serialization as se
from hippolyzer.lib.base.network.transport import Direction, SocketUDPTransport
from hippolyzer.lib.proxy.addons import BaseInteractionManager, AddonManager
from hippolyzer.lib.proxy.ca_utils import setup_ca_everywhere
from hippolyzer.lib.proxy.caps_client import ProxyCapsClient
from hippolyzer.lib.proxy.http_proxy import create_proxy_master, HTTPFlowContext
from hippolyzer.lib.proxy.message_logger import LLUDPMessageLogEntry, AbstractMessageLogEntry, WrappingMessageLogger, \
    import_log_entries, export_log_entries
from hippolyzer.lib.proxy.region import ProxiedRegion
from hippolyzer.lib.proxy.sessions import Session, SessionManager
from hippolyzer.lib.proxy.settings import ProxySettings
from hippolyzer.lib.proxy.templates import CAP_TEMPLATES

LOG = logging.getLogger(__name__)

MAIN_WINDOW_UI_PATH = get_resource_filename("apps/proxy_mainwindow.ui")
MESSAGE_BUILDER_UI_PATH = get_resource_filename("apps/message_builder.ui")
ADDON_DIALOG_UI_PATH = get_resource_filename("apps/addon_dialog.ui")
FILTER_DIALOG_UI_PATH = get_resource_filename("apps/filter_dialog.ui")


def show_error_message(error_msg, parent=None):
    error_dialog = QtWidgets.QErrorMessage(parent=parent)
    # No obvious way to set this to plaintext, yuck...
    error_dialog.showMessage(html.escape(error_msg))
    error_dialog.exec()
    error_dialog.raise_()


class GUISessionManager(SessionManager, QtCore.QObject):
    regionAdded = QtCore.Signal(ProxiedRegion)
    regionRemoved = QtCore.Signal(ProxiedRegion)

    def __init__(self, settings):
        SessionManager.__init__(self, settings)
        QtCore.QObject.__init__(self)
        self.all_regions = []
        self.message_logger = WrappingMessageLogger()

    def checkRegions(self):
        new_regions = itertools.chain(*[s.regions for s in self.sessions])
        new_regions = [r for r in new_regions if r.is_alive]
        for new_region in new_regions:
            if new_region not in self.all_regions:
                self.regionAdded.emit(new_region)  # type: ignore
        for old_region in self.all_regions:
            if old_region not in new_regions:
                self.regionRemoved.emit(old_region)  # type: ignore

        self.all_regions = new_regions


class GUIInteractionManager(BaseInteractionManager):
    def __init__(self, parent: QtWidgets.QWidget):
        BaseInteractionManager.__init__(self)
        self._parent = parent

    def main_window_handle(self) -> Any:
        return self._parent

    def _dialog_async_exec(self, dialog: QtWidgets.QDialog):
        future = asyncio.Future()
        dialog.finished.connect(lambda r: future.set_result(r))
        dialog.open()
        return future

    async def _file_dialog(
            self, caption: str, directory: str, filter_str: str, mode: QtWidgets.QFileDialog.FileMode,
            default_suffix: str = '',
    ) -> Tuple[bool, QtWidgets.QFileDialog]:
        dialog = QtWidgets.QFileDialog(self._parent, caption=caption, directory=directory, filter=filter_str)
        dialog.setFileMode(mode)
        if mode == QtWidgets.QFileDialog.FileMode.AnyFile:
            dialog.setAcceptMode(QtWidgets.QFileDialog.AcceptMode.AcceptSave)
        if default_suffix:
            dialog.setDefaultSuffix(default_suffix)
        res = await self._dialog_async_exec(dialog)
        return res, dialog

    async def open_files(self, caption: str = '', directory: str = '', filter_str: str = '') -> List[str]:
        res, dialog = await self._file_dialog(
            caption, directory, filter_str, QtWidgets.QFileDialog.FileMode.ExistingFiles
        )
        if not res:
            return []
        return dialog.selectedFiles()

    async def open_file(self, caption: str = '', directory: str = '', filter_str: str = '') -> Optional[str]:
        res, dialog = await self._file_dialog(
            caption, directory, filter_str, QtWidgets.QFileDialog.FileMode.ExistingFile
        )
        if not res:
            return None
        return dialog.selectedFiles()[0]

    async def open_dir(self, caption: str = '', directory: str = '', filter_str: str = '') -> Optional[str]:
        res, dialog = await self._file_dialog(
            caption, directory, filter_str, QtWidgets.QFileDialog.FileMode.Directory
        )
        if not res:
            return None
        return dialog.selectedFiles()[0]

    async def save_file(self, caption: str = '', directory: str = '', filter_str: str = '',
                        default_suffix: str = '') -> Optional[str]:
        res, dialog = await self._file_dialog(
            caption, directory, filter_str, QtWidgets.QFileDialog.FileMode.AnyFile, default_suffix,
        )
        if not res or not dialog.selectedFiles():
            return None
        return dialog.selectedFiles()[0]

    async def confirm(self, title: str, caption: str) -> bool:
        msg = QtWidgets.QMessageBox(
            QtWidgets.QMessageBox.Icon.Question,
            title,
            caption,
            QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel,
            self._parent,
        )
        fut = asyncio.Future()
        msg.finished.connect(lambda r: fut.set_result(r))
        msg.open()
        return (await fut) == QtWidgets.QMessageBox.Ok


class GUIProxySettings(ProxySettings):
    FIRST_RUN: bool = SettingDescriptor(True)

    """Persistent settings backed by QSettings"""
    def __init__(self, settings: QtCore.QSettings):
        super().__init__()
        self._settings_obj = settings

    def get_setting(self, name: str) -> Any:
        val: Any = self._settings_obj.value(name, defaultValue=dataclasses.MISSING)
        if val is dataclasses.MISSING:
            return val
        return json.loads(val)

    def set_setting(self, name: str, val: Any):
        self._settings_obj.setValue(name, json.dumps(val))


def nonFatalExceptions(f):
    @functools.wraps(f)
    def _wrapper(self, *args, **kwargs):
        try:
            return f(self, *args, **kwargs)
        except Exception as e:
            logging.exception("Treating exception as non-fatal")
            show_error_message(str(e), parent=self)
            return None

    return _wrapper


def buildReplacements(session: Session, region: ProxiedRegion):
    if not session or not region:
        return {}
    selected = session.selected
    agent_object = region.objects.lookup_fullid(session.agent_id)
    selected_local = selected.object_local
    selected_object = None
    if selected_local:
        # We may or may not have an object for this
        selected_object = region.objects.lookup_localid(selected_local)
    return {
        "SELECTED_LOCAL": selected_local,
        "SELECTED_FULL": selected_object.FullID if selected_object else None,
        "SELECTED_PARCEL_LOCAL": selected.parcel_local,
        "SELECTED_PARCEL_FULL": selected.parcel_full,
        "SELECTED_SCRIPT_ITEM": selected.script_item,
        "SELECTED_TASK_ITEM": selected.task_item,
        "AGENT_ID": session.agent_id,
        "AGENT_LOCAL": agent_object.LocalID if agent_object else None,
        "SESSION_ID": session.id,
        "AGENT_POS": agent_object.Position if agent_object else None,
        "NULL_KEY": UUID(),
        "RANDOM_KEY": UUID.random,
        "CIRCUIT_CODE": session.circuit_code,
        "REGION_HANDLE": region.handle,
    }


class MessageLogWindow(QtWidgets.QMainWindow):
    DEFAULT_IGNORE = "StartPingCheck CompletePingCheck PacketAck SimulatorViewerTimeMessage SimStats " \
                     "AgentUpdate AgentAnimation AvatarAnimation ViewerEffect CoarseLocationUpdate LayerData " \
                     "CameraConstraint ObjectUpdateCached RequestMultipleObjects ObjectUpdate ObjectUpdateCompressed " \
                     "ImprovedTerseObjectUpdate KillObject ImagePacket SoundTrigger AttachedSound PreloadSound " \
                     "GetMesh GetMesh2 EventQueueGet RequestObjectPropertiesFamily ObjectPropertiesFamily " \
                     "AvatarRenderInfo FirestormBridge ObjectAnimation ParcelDwellRequest ParcelAccessListRequest " \
                     "ParcelDwellReply ParcelAccessListReply AttachedSoundGainChange " \
                     "ParcelPropertiesRequest ParcelProperties GetObjectCost GetObjectPhysicsData ObjectImage " \
                     "ViewerAsset GetTexture SetAlwaysRun GetDisplayNames MapImageService MapItemReply".split(" ")
    DEFAULT_FILTER = f"!({' || '.join(ignored for ignored in DEFAULT_IGNORE)})"

    textRequest: QtWidgets.QTextEdit

    def __init__(
            self, settings: GUIProxySettings, session_manager: GUISessionManager,
            log_live_messages: bool, parent: Optional[QtWidgets.QWidget] = None,
    ):
        super().__init__(parent=parent)
        loadUi(MAIN_WINDOW_UI_PATH, self)

        if parent:
            self.setWindowTitle("Message Log")
            self.menuBar.setEnabled(False)  # type: ignore
            self.menuBar.hide()  # type: ignore

        self._selectedEntry: Optional[AbstractMessageLogEntry] = None

        self.settings = settings
        self.sessionManager = session_manager
        if log_live_messages:
            self.model = MessageLogModel(parent=self.tableView)
            session_manager.message_logger.loggers.append(self.model)
        else:
            self.model = MessageLogModel(parent=self.tableView, maxlen=None)
        self.tableView.setModel(self.model)
        self.model.rowsAboutToBeInserted.connect(self.beforeInsert)
        self.model.rowsInserted.connect(self.afterInsert)
        self.tableView.selectionModel().selectionChanged.connect(self._messageSelected)
        self.checkBeautify.clicked.connect(self._showSelectedMessage)
        self.checkPause.clicked.connect(self._setPaused)
        self.setFilter(self.DEFAULT_FILTER)
        self.btnClearLog.clicked.connect(self.model.clear)
        self.lineEditFilter.editingFinished.connect(self.setFilter)
        self.btnMessageBuilder.clicked.connect(self._sendToMessageBuilder)
        self.btnCopyRepr.clicked.connect(self._copyRepr)
        self.actionInstallHTTPSCerts.triggered.connect(self.installHTTPSCerts)
        self.actionManageAddons.triggered.connect(self._manageAddons)
        self.actionManageFilters.triggered.connect(self._manageFilters)
        self.actionOpenMessageBuilder.triggered.connect(self._openMessageBuilder)

        self.actionProxyRemotelyAccessible.setChecked(self.settings.REMOTELY_ACCESSIBLE)
        self.actionUseViewerObjectCache.setChecked(self.settings.USE_VIEWER_OBJECT_CACHE)
        self.actionRequestMissingObjects.setChecked(self.settings.AUTOMATICALLY_REQUEST_MISSING_OBJECTS)
        self.actionProxyRemotelyAccessible.triggered.connect(self._setProxyRemotelyAccessible)
        self.actionUseViewerObjectCache.triggered.connect(self._setUseViewerObjectCache)
        self.actionRequestMissingObjects.triggered.connect(self._setRequestMissingObjects)
        self.actionOpenNewMessageLogWindow.triggered.connect(self._openNewMessageLogWindow)
        self.actionImportLogEntries.triggered.connect(self._importLogEntries)
        self.actionExportLogEntries.triggered.connect(self._exportLogEntries)

        self._filterMenu = QtWidgets.QMenu()
        self._populateFilterMenu()
        self.toolButtonFilter.setMenu(self._filterMenu)

        self._shouldScrollOnInsert = True
        self.tableView.horizontalHeader().resizeSection(MessageLogHeader.Host, 80)
        self.tableView.horizontalHeader().resizeSection(MessageLogHeader.Method, 60)
        self.tableView.horizontalHeader().resizeSection(MessageLogHeader.Name, 180)
        self.tableView.horizontalHeader().setStretchLastSection(True)

        self.textResponse.hide()

    def closeEvent(self, event) -> None:
        loggers = self.sessionManager.message_logger.loggers
        if self.model in loggers:
            loggers.remove(self.model)
        super().closeEvent(event)

    def _populateFilterMenu(self):
        def _addFilterAction(text, filter_str):
            filter_action = QtGui.QAction(text, self)
            filter_action.triggered.connect(lambda: self.setFilter(filter_str))
            self._filterMenu.addAction(filter_action)

        self._filterMenu.clear()

        _addFilterAction("Default", self.DEFAULT_FILTER)
        filters = self.settings.FILTERS
        for preset_name, preset_filter in filters.items():
            _addFilterAction(preset_name, preset_filter)

    def getFilterDict(self):
        return self.settings.FILTERS

    def setFilterDict(self, val: dict):
        self.settings.FILTERS = val
        self._populateFilterMenu()

    def _manageFilters(self):
        dialog = FilterDialog(self)
        dialog.exec()

    @nonFatalExceptions
    def setFilter(self, filter_str=None):
        if filter_str is None:
            filter_str = self.lineEditFilter.text()
        else:
            self.lineEditFilter.setText(filter_str)
        self.model.set_filter(filter_str)

    def _setPaused(self, checked):
        self.model.set_paused(checked)

    def _messageSelected(self, selected, _deselected):
        indexes = selected.indexes()
        if not len(indexes):
            self.btnMessageBuilder.setEnabled(False)
            self.btnCopyRepr.setEnabled(False)
            return
        index = indexes[0]
        entry: AbstractMessageLogEntry = index.data(QtCore.Qt.UserRole)
        self._selectedEntry = entry
        self.btnMessageBuilder.setEnabled(True)
        if isinstance(entry, LLUDPMessageLogEntry):
            self.btnCopyRepr.setEnabled(True)
        else:
            self.btnCopyRepr.setEnabled(False)
        self._showSelectedMessage()

    def _showSelectedMessage(self):
        entry = self._selectedEntry
        if not entry:
            return
        req = entry.request(
            beautify=self.checkBeautify.isChecked(),
            replacements=buildReplacements(entry.session, entry.region),
        )
        self.textRequest.setPlainText(req)
        # The string has a map of fields and their associated positions within the string,
        # use that to highlight any individual fields the filter matched on.
        if isinstance(req, SpannedString):
            for field in self.model.filter.match(entry, short_circuit=False).fields:
                field_span = req.spans.get(field)
                if not field_span:
                    continue
                cursor = self.textRequest.textCursor()
                cursor.setPosition(field_span[0], QtGui.QTextCursor.MoveAnchor)
                cursor.setPosition(field_span[1], QtGui.QTextCursor.KeepAnchor)
                highlight_format = QtGui.QTextBlockFormat()
                highlight_format.setBackground(QtCore.Qt.yellow)
                cursor.setBlockFormat(highlight_format)

        resp = entry.response(beautify=self.checkBeautify.isChecked())
        if resp:
            self.textResponse.show()
            self.textResponse.setPlainText(resp)
        else:
            self.textResponse.hide()

    def beforeInsert(self):
        vbar = self.tableView.verticalScrollBar()
        self._shouldScrollOnInsert = vbar.value() == vbar.maximum()

    def afterInsert(self):
        if self._shouldScrollOnInsert:
            self.tableView.scrollToBottom()

    def _sendToMessageBuilder(self):
        if not self._selectedEntry:
            return
        win = MessageBuilderWindow(self, self.sessionManager)
        win.show()
        msg = self._selectedEntry
        beautify = self.checkBeautify.isChecked()
        replacements = buildReplacements(msg.session, msg.region)
        win.setMessageText(msg.request(beautify=beautify, replacements=replacements))

    @nonFatalExceptions
    def _copyRepr(self):
        if not self._selectedEntry:
            return
        if not isinstance(self._selectedEntry, LLUDPMessageLogEntry):
            return
        msg_repr = self._selectedEntry.message.repr(pretty=True)
        QtGui.QGuiApplication.clipboard().setText(msg_repr)

    def _openMessageBuilder(self):
        win = MessageBuilderWindow(self, self.sessionManager)
        win.show()

    def _openNewMessageLogWindow(self):
        win: QtWidgets.QMainWindow = MessageLogWindow(
            self.settings, self.sessionManager, log_live_messages=True, parent=self)
        win.setFilter(self.lineEditFilter.text())
        win.show()
        win.activateWindow()

    @asyncSlot()
    async def _importLogEntries(self):
        log_file = await AddonManager.UI.open_file(
            caption="Import Log Entries", filter_str="Hippolyzer Logs (*.hippolog)"
        )
        if not log_file:
            return
        win = MessageLogWindow(self.settings, self.sessionManager, log_live_messages=False, parent=self)
        win.setFilter(self.lineEditFilter.text())
        with open(log_file, "rb") as f:
            entries = import_log_entries(f.read())
        for entry in entries:
            win.model.add_log_entry(entry)
        win.show()
        win.activateWindow()

    @asyncSlot()
    async def _exportLogEntries(self):
        log_file = await AddonManager.UI.save_file(
            caption="Export Log Entries", filter_str="Hippolyzer Logs (*.hippolog)", default_suffix="hippolog",
        )
        if not log_file:
            return
        with open(log_file, "wb") as f:
            f.write(export_log_entries(self.model))

    def installHTTPSCerts(self):
        msg = QtWidgets.QMessageBox()
        msg.setText("Would you like to install the proxy's HTTPS certificate in the config dir"
                    " of any installed viewers so that HTTPS connections will work?")
        yes_btn = msg.addButton("Yes", QtWidgets.QMessageBox.NoRole)
        msg.addButton("No", QtWidgets.QMessageBox.NoRole)
        msg.exec()
        clicked_btn = msg.clickedButton()
        if clicked_btn is not yes_btn:
            return

        master = create_proxy_master("127.0.0.1", -1, HTTPFlowContext())
        dirs = setup_ca_everywhere(master)

        msg = QtWidgets.QMessageBox()
        if dirs:
            msg.setText(f"Installed certificate in {[str(x) for x in dirs]!r}, restart your viewer!")
        else:
            msg.setText("Couldn't find any viewer config directories!")
        msg.exec()

    def _setProxyRemotelyAccessible(self, checked: bool):
        self.sessionManager.settings.REMOTELY_ACCESSIBLE = checked
        msg = QtWidgets.QMessageBox()
        msg.setText("Remote accessibility setting changes will take effect on next run")
        msg.exec()

    def _setUseViewerObjectCache(self, checked: bool):
        self.sessionManager.settings.USE_VIEWER_OBJECT_CACHE = checked

    def _setRequestMissingObjects(self, checked: bool):
        self.sessionManager.settings.AUTOMATICALLY_REQUEST_MISSING_OBJECTS = checked

    def _manageAddons(self):
        dialog = AddonDialog(self)
        dialog.exec()

    def getAddonList(self) -> List[str]:
        return self.sessionManager.settings.ADDON_SCRIPTS

    def setAddonList(self, val: List[str]):
        self.sessionManager.settings.ADDON_SCRIPTS = val


BANNED_HEADERS = ("content-length", "host")


def _parse_http_message(msg_text):
    request_line, mime_msg = re.split(r"\r?\n", msg_text, 1)
    method, uri, version = msg_text.split(" ", 2)
    # comment line for URI, throw it out.
    while mime_msg.startswith("#"):
        _, mime_msg = re.split(r"\r?\n", mime_msg, 1)
    msg = email.message_from_string(mime_msg)
    # Throw out Content-Length, very likely to be wrong after beautification
    headers = multidict.CIMultiDict(
        (k, v) for k, v in msg.items() if k.lower() not in BANNED_HEADERS
    )
    # Use to get the body rather than `email.` since that wants to do some
    # fancy parsing of MIME bits that we don't want it to do.
    msg_parts = re.split(r"\r?\n\r?\n", msg_text, 1)
    body = None
    if len(msg_parts) == 2:
        body = msg_parts[1]
    return method, uri, headers, body


def _coerce_to_bytes(val):
    if isinstance(val, bytes):
        return val
    if not isinstance(val, str):
        val = str(val)
    return val.encode("utf8")


class MessageBuilderWindow(QtWidgets.QMainWindow):
    def __init__(self, parent, session_manager):
        super().__init__(parent=parent)
        loadUi(MESSAGE_BUILDER_UI_PATH, self)
        self.templateDict = DEFAULT_TEMPLATE_DICT
        self.llsdSerializer = LLSDMessageSerializer()
        self.sessionManager: SessionManager = session_manager
        self.regionModel = RegionListModel(self, self.sessionManager)
        self.listRegions.setModel(self.regionModel)

        self._populateMessageTypeMenus()
        self.comboTrusted.textActivated.connect(self._fillUDPMessage)
        self.comboUntrusted.textActivated.connect(self._fillUDPMessage)
        self.comboCaps.textActivated.connect(self._fillCapsMessage)

        self.btnSend.clicked.connect(self._sendMessage)
        self.textRequest.setFocus(QtCore.Qt.FocusReason.OtherFocusReason)

    def _getTarget(self):
        region_indexes = self.listRegions.selectionModel().selectedIndexes()
        session, region = None, None
        if region_indexes:
            region = region_indexes[0].data(QtCore.Qt.UserRole)
            session = region.session()
        elif self.sessionManager.sessions:
            session = self.sessionManager.sessions[-1]
            region = session.main_region
        return session, region

    def setMessageText(self, text):
        self.textRequest.setPlainText(text)

    def _populateMessageTypeMenus(self):
        self.comboTrusted.clear()
        self.comboUntrusted.clear()
        self.comboCaps.clear()

        self.comboTrusted.addItem("<Trusted>")
        self.comboUntrusted.addItem("<Untrusted>")
        self.comboCaps.addItem("<Caps>")

        message_names = sorted(x.name for x in self.templateDict)

        for message_name in message_names:
            if self.templateDict[message_name].msg_trust:
                self.comboTrusted.addItem(message_name)
            else:
                self.comboUntrusted.addItem(message_name)

        cap_names = sorted(set(itertools.chain(*[r.cap_urls.keys() for r in self.regionModel.regions])))
        for cap_name in cap_names:
            if cap_name.endswith("ProxyWrapper"):
                continue
            self.comboCaps.addItem(cap_name)

    def _fillCapsMessage(self, cap_name):
        if cap_name.startswith("<"):
            return
        session, region = self._getTarget()
        self.textRequest.clear()
        params = ""
        path = ""
        body = "<llsd><undef/></llsd>"
        method = "POST"
        headers = "Content-Type: application/llsd+xml\nX-Hippo-Directives: 1\n"
        for cap_template in CAP_TEMPLATES:
            # Might be nice to allow selecting method variants?
            if cap_template.cap_name == cap_name:
                params = urllib.parse.urlencode({x: '' for x in cap_template.query})
                if params:
                    params = '?' + params
                body = cap_template.body.decode("utf8")
                path = cap_template.path
                method = cap_template.method
                if method == "GET":
                    # No content-type because no content
                    headers = ""
                break
        self.textRequest.setPlainText(
            f"""{method} [[{cap_name}]]{path}{params} HTTP/1.1
# {region.cap_urls.get(cap_name, "<unknown URI>")}
{headers}
{body}"""
        )

    def _fillUDPMessage(self, message_name):
        if message_name.startswith("<"):
            return
        self.textRequest.clear()

        template = self.templateDict[message_name]
        msg = Message(message_name, direction=Direction.OUT)

        for tmpl_block in template.blocks:
            num_blocks = tmpl_block.number or 1
            for i in range(num_blocks):
                fill_vars = {}
                for var in tmpl_block.variables:
                    fill_vars[var.name] = self._getVarPlaceholder(template, tmpl_block, var)
                msg_block = Block(tmpl_block.name, **fill_vars)
                msg.add_block(msg_block)
        self.textRequest.setPlainText(
            HumanMessageSerializer.to_human_string(msg, replacements={}, beautify=True, template=template)
        )

    def _getVarPlaceholder(self, msg, block, var):
        if block.name == "AgentData":
            if var.name == "AgentID":
                return VerbatimHumanVal("[[AGENT_ID]]")
            if var.name == "SessionID":
                return VerbatimHumanVal("[[SESSION_ID]]")
        if block.name == "ObjectData":
            if var.name.endswith("ID") and var.name != "AgentID":
                if var.type == MsgType.MVT_LLUUID:
                    return VerbatimHumanVal("[[SELECTED_FULL]]")
                else:
                    return VerbatimHumanVal("[[SELECTED_LOCAL]]")
        if "Parcel" in msg.name:
            if var.name.endswith("LocalID"):
                return VerbatimHumanVal("[[SELECTED_PARCEL_LOCAL]]")
        if "ParcelID" in var.name:
            return VerbatimHumanVal("[[SELECTED_PARCEL_FULL]]")
        if "Handle" in var.name:
            return VerbatimHumanVal("[[REGION_HANDLE]]")
        if "CircuitCode" in var.name or ("Circuit" in block.name and "Code" in var.name):
            return VerbatimHumanVal("[[CIRCUIT_CODE]]")
        if var.name.endswith("LocalID"):
            return VerbatimHumanVal("[[SELECTED_LOCAL]]")
        if any(var.name.endswith(x) for x in ("TransactionID", "TransferID", "Invoice", "QueryID")):
            return VerbatimHumanVal("[[RANDOM_KEY]]")
        if var.name in ("TaskID", "ObjectID"):
            return VerbatimHumanVal("[[SELECTED_FULL]]")

        default_val = var.default_value
        if default_val is not None:
            return default_val
        return VerbatimHumanVal("")

    @nonFatalExceptions
    def _sendMessage(self, _checked=False):
        session, region = self._getTarget()

        msg_text = self.textRequest.toPlainText()
        replacements = buildReplacements(session, region)

        if re.match(r"\A\s*(in|out)\s+", msg_text, re.I):
            sender_func = self._sendLLUDPMessage
        elif re.match(r"\A\s*(eq)\s+", msg_text, re.I):
            sender_func = self._sendEQMessage
        elif re.match(r"\A.*http/[0-9.]+\r?\n", msg_text, re.I):
            sender_func = self._sendHTTPMessage
        else:
            raise ValueError("IDK what kind of message %s is" % msg_text)

        for i in range(0, self.spinRepeat.value() + 1):
            sender_func(session, region, msg_text, {"i": i, **replacements})

    def _buildEnv(self, session, region):
        env = {}
        if region:
            env["full_to_local"] = lambda x: region.objects.lookup_fullid(x).LocalID
            env["region"] = region
        if session:
            env["session"] = session
        return env

    def _sendLLUDPMessage(self, session, region: Optional[ProxiedRegion], msg_text, replacements):
        if not session or not region:
            raise RuntimeError("Need a valid session and region to send LLUDP")
        env = self._buildEnv(session, region)
        # We specifically want to allow `eval()` in messages since
        # messages from here are trusted.
        msg = HumanMessageSerializer.from_human_string(msg_text, replacements, env, safe=False)
        if self.checkLLUDPViaCaps.isChecked():
            if msg.direction == Direction.IN:
                region.eq_manager.inject_message(msg)
            else:
                self._sendHTTPRequest(
                    "POST",
                    region.cap_urls["UntrustedSimulatorMessage"],
                    {"Content-Type": "application/llsd+xml", "Accept": "application/llsd+xml"},
                    self.llsdSerializer.serialize(msg),
                )
        else:
            transport = None
            off_circuit = self.checkOffCircuit.isChecked()
            if off_circuit:
                transport = SocketUDPTransport(socket.socket(socket.AF_INET, socket.SOCK_DGRAM))
            region.circuit.send(msg, transport=transport)
            if off_circuit:
                transport.close()

    def _sendEQMessage(self, session, region: Optional[ProxiedRegion], msg_text: str, replacements: dict):
        if not session or not region:
            raise RuntimeError("Need a valid session and region to send EQ event")
        message_line, _, body = (x.strip() for x in msg_text.partition("\n"))
        message_name = message_line.rsplit(" ", 1)[-1]

        env = self._buildEnv(session, region)

        def directive_handler(m):
            return self._handleHTTPDirective(env, replacements, False, m)
        body = re.sub(rb"<!HIPPO(\w+)\[\[(.*?)]]>", directive_handler, body.encode("utf8"), flags=re.S)

        region.eq_manager.inject_event({
            "message": message_name,
            "body": llsd.parse_xml(body),
        })

    def _sendHTTPMessage(self, session, region, msg_text: str, replacements: dict):
        env = self._buildEnv(session, region)
        method, uri, headers, body = _parse_http_message(msg_text)

        has_directives = headers.get("X-Hippo-Directives") == "1"
        is_escaped = headers.get("X-Hippo-Escaped-Body") == "1"
        is_beautified = headers.get("X-Hippo-Beautify") == "1"

        if body:
            body = body.encode("utf8")
            # Only handle directives if specifically asked for
            if has_directives:
                def directive_handler(m):
                    return self._handleHTTPDirective(env, replacements, is_escaped, m)
                body = re.sub(rb"<!HIPPO(\w+)\[\[(.*?)]]>", directive_handler, body, flags=re.S)
        match = re.match(r"\A\[\[(.*?)]](.*)", uri)
        cap_name = None
        if match:
            cap_name = match.group(1)
            cap_url = session.global_caps.get(cap_name)
            if not cap_url:
                cap_url = region.cap_urls.get(cap_name)
            if not cap_url:
                raise ValueError("Don't have a Cap for %s" % cap_name)
            uri = cap_url + match.group(2)

        # The body contains python escape sequences, decode them.
        if is_escaped:
            body = bytes_unescape(body)

        # Convert from human-friendly representation before sending
        if cap_name and is_beautified:
            parsed = llsd.parse_xml(body)
            serializer = se.HTTP_SERIALIZERS.get(cap_name)
            if serializer:
                body = serializer.serialize_req_body(method, parsed)
                if body is se.UNSERIALIZABLE:
                    raise ValueError(f"Cannot serialize {parsed!r}")

        headers.pop("X-Hippo-Beautify", None)
        headers.pop("X-Hippo-Directives", None)
        headers.pop("X-Hippo-Escaped-Body", None)
        self._sendHTTPRequest(method, uri, headers, body)

    def _handleHTTPDirective(self, env: dict, replacements: dict, need_escaped: bool, match: re.Match) -> bytes:
        # HTTP messages from the builder can have inline directives like
        # <!HIPPOBASE64[[foobar]]>, handle those.
        directive: bytes = match.group(1)
        contents: bytes = match.group(2)
        unescaped_contents: bytes = bytes_unescape(contents)

        if directive == b"BASE64":
            val = base64.b64encode(unescaped_contents)
        elif directive == b"CDATA":
            val = contents.replace(b"&", b"&amp;").replace(b"<", b"&lt;").replace(b">", b"&gt;")
        elif directive == b"UNESCAPE":
            val = unescaped_contents
        elif directive == b"EVAL":
            val = subfield_eval(contents.decode("utf8").strip(), globals_={**env, **replacements})
            val = _coerce_to_bytes(val)
        elif directive == b"REPL":
            repl = replacements[contents.decode("utf8").strip()]
            if callable(repl):
                repl = repl()
            val = _coerce_to_bytes(repl)
        else:
            raise ValueError(f"Unknown directive {directive}")

        # We're going to output this in a context that expects string escape sequences
        # This isn't very nice, but saves me writing a parser for the language in HTTP
        # request bodies.
        if need_escaped:
            return bytes_escape(val)
        return val

    def _sendHTTPRequest(self, method, uri, headers, body):
        caps_client = ProxyCapsClient(self.sessionManager.settings)

        async def _send_request():
            req = caps_client.request(method, uri, headers=headers, data=body)
            async with req as resp:
                # Read, but throw away the resp so the connection is kept alive long
                # enough for the full response to pass through the proxy
                await resp.read()

        asyncio.create_task(_send_request())


class AddonDialog(QtWidgets.QDialog):
    listAddons: QtWidgets.QListWidget

    def __init__(self, parent: MessageLogWindow):
        super().__init__()

        loadUi(ADDON_DIALOG_UI_PATH, self)
        self._parent = parent
        self.btnLoad.pressed.connect(self._loadPressed)
        self.btnRemove.pressed.connect(self._removeCurrentPressed)
        self._populateAddons()

    def _populateAddons(self):
        for addon in self._parent.getAddonList():
            self.listAddons.addItem(QtWidgets.QListWidgetItem(addon))

    @nonFatalExceptions
    def _removeCurrentPressed(self):
        selected = self.listAddons.selectedItems()
        if selected:
            item: QtWidgets.QListWidgetItem = selected[0]
            try:
                AddonManager.unload_addon_from_path(item.text(), reload=True)
            except Exception as e:
                # This addon might not even really be loaded
                show_error_message(error_msg=str(e), parent=self)
            addons = self._parent.getAddonList()
            addons.remove(item.text())
            self._parent.setAddonList(addons)
            idx = self.listAddons.indexFromItem(item).row()
            self.listAddons.takeItem(idx)

    @nonFatalExceptions
    def _loadPressed(self):
        file_filter = 'Python Addon (*.py)'

        options = QtWidgets.QFileDialog.Options()
        path, _ = QtWidgets.QFileDialog.getOpenFileName(self, 'Load', '', file_filter, options=options)
        if not path:
            return
        addons = self._parent.getAddonList()
        path = str(pathlib.Path(path).absolute())
        # already loaded
        if path in addons:
            return
        AddonManager.load_addon_from_path(path, reload=True)
        addons.append(path)
        self.listAddons.addItem(QtWidgets.QListWidgetItem(path))
        self._parent.setAddonList(addons)


class FilterDialog(QtWidgets.QDialog):
    listFilters: QtWidgets.QListWidget

    def __init__(self, parent: MessageLogWindow):
        super().__init__()

        loadUi(FILTER_DIALOG_UI_PATH, self)
        self._parent = parent
        self.btnSaveCurrent.pressed.connect(self._saveCurrent)
        self.btnRemove.pressed.connect(self._removeCurrentPressed)
        self._populateFilters()

    def _populateFilters(self):
        self.listFilters.clear()
        for filter_obj in self._parent.getFilterDict():
            self.listFilters.addItem(QtWidgets.QListWidgetItem(filter_obj))

    @nonFatalExceptions
    def _saveCurrent(self):
        filters = self._parent.getFilterDict()
        filter_text = self._parent.lineEditFilter.text()
        name, ok = QtWidgets.QInputDialog.getText(self, "Filter Name", "Filter Name",
                                                  QtWidgets.QLineEdit.EchoMode.Normal)
        if not name or not ok:
            return
        filters[name] = filter_text
        self._parent.setFilterDict(filters)
        self._populateFilters()

    @nonFatalExceptions
    def _removeCurrentPressed(self):
        selected = self.listFilters.selectedItems()
        if selected:
            item: QtWidgets.QListWidgetItem = selected[0]
            filters = self._parent.getFilterDict()
            del filters[item.text()]
            self._parent.setFilterDict(filters)
            idx = self.listFilters.indexFromItem(item).row()
            self.listFilters.takeItem(idx)


def gui_main():
    multiprocessing.set_start_method('spawn')
    QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_ShareOpenGLContexts)
    app = QtWidgets.QApplication(sys.argv)
    loop = QEventLoop(app)
    asyncio.set_event_loop(loop)
    settings = GUIProxySettings(QtCore.QSettings("SaladDais", "hippolyzer"))
    session_manager = GUISessionManager(settings)
    window = MessageLogWindow(settings, session_manager, log_live_messages=True)
    AddonManager.UI = GUIInteractionManager(window)
    timer = QtCore.QTimer(app)
    timer.timeout.connect(window.sessionManager.checkRegions)
    timer.start(100)
    signal.signal(signal.SIGINT, lambda *args: QtWidgets.QApplication.quit())
    window.show()
    http_host = None
    if window.sessionManager.settings.REMOTELY_ACCESSIBLE:
        http_host = "0.0.0.0"
    if settings.FIRST_RUN:
        settings.FIRST_RUN = False
        # Automatically offer to install the HTTPS certs on first run.
        window.installHTTPSCerts()
    start_proxy(
        session_manager=window.sessionManager,
        extra_addon_paths=window.getAddonList(),
        proxy_host=http_host,
    )


if __name__ == "__main__":
    multiprocessing.freeze_support()
    gui_main()



"""
Manager for a client's view of objects in the region and world.
"""

from __future__ import annotations

import asyncio
import collections
import enum
import itertools
import logging
import math
import weakref
from typing import *

from hippolyzer.lib.base.datatypes import UUID, Vector3
from hippolyzer.lib.base.helpers import proxify
from hippolyzer.lib.base.message.message import Block, Message
from hippolyzer.lib.base.message.message_handler import MessageHandler
from hippolyzer.lib.base.message.msgtypes import PacketFlags
from hippolyzer.lib.base.objects import (
    normalize_object_update,
    normalize_terse_object_update,
    normalize_object_update_compressed_data,
    normalize_object_update_compressed,
    Object, handle_to_global_pos,
)
from hippolyzer.lib.base.settings import Settings
from hippolyzer.lib.client.namecache import NameCache, NameCacheEntry
from hippolyzer.lib.client.state import BaseClientSession, BaseClientRegion
from hippolyzer.lib.base.templates import PCode, ObjectStateSerializer


LOG = logging.getLogger(__name__)
OBJECT_OR_LOCAL = Union[Object, int]


class ObjectUpdateType(enum.IntEnum):
    OBJECT_UPDATE = enum.auto()
    PROPERTIES = enum.auto()
    FAMILY = enum.auto()
    COSTS = enum.auto()
    KILL = enum.auto()


class ClientObjectManager:
    """
    Object manager for a specific region
    """

    __slots__ = ("_region", "_world_objects", "state")

    def __init__(self, region: BaseClientRegion):
        self._region: BaseClientRegion = proxify(region)
        self._world_objects: ClientWorldObjectManager = proxify(region.session().objects)
        self.state: RegionObjectsState = RegionObjectsState()

    def __len__(self):
        return len(self.state.localid_lookup)

    @property
    def all_objects(self) -> Iterable[Object]:
        return self.state.localid_lookup.values()

    @property
    def missing_locals(self) -> Set[int]:
        return self.state.missing_locals

    def clear(self):
        self.state.clear()
        if self._region.handle is not None:
            # We're tracked by the world object manager, tell it to untrack
            # any objects that we owned
            self._world_objects.clear_region_objects(self._region.handle)

    def lookup_localid(self, localid: int) -> Optional[Object]:
        return self.state.lookup_localid(localid)

    def lookup_fullid(self, fullid: UUID) -> Optional[Object]:
        obj = self._world_objects.lookup_fullid(fullid)
        if obj is None or obj.RegionHandle != self._region.handle:
            return None
        return obj

    @property
    def all_avatars(self) -> Iterable[Avatar]:
        return tuple(a for a in self._world_objects.all_avatars
                     if a.RegionHandle == self._region.handle)

    def lookup_avatar(self, fullid: UUID) -> Optional[Avatar]:
        for avatar in self.all_avatars:
            if avatar.FullID == fullid:
                return avatar
        return None

    # noinspection PyUnusedLocal
    def _is_localid_selected(self, local_id: int):
        return False

    def request_object_properties(self, objects: Union[OBJECT_OR_LOCAL, Sequence[OBJECT_OR_LOCAL]]) \
            -> List[asyncio.Future[Object]]:
        if isinstance(objects, (Object, int)):
            objects = (objects,)
        if not objects:
            return []

        local_ids = tuple((o.LocalID if isinstance(o, Object) else o) for o in objects)

        # Don't mess with already selected objects
        unselected_ids = tuple(local for local in local_ids if not self._is_localid_selected(local))
        ids_to_req = unselected_ids

        session = self._region.session()
        while ids_to_req:
            blocks = [
                Block("AgentData", AgentID=session.agent_id, SessionID=session.id),
                *[Block("ObjectData", ObjectLocalID=x) for x in ids_to_req[:255]],
            ]
            # Selecting causes ObjectProperties to be sent
            self._region.circuit.send(Message("ObjectSelect", blocks, flags=PacketFlags.RELIABLE))
            self._region.circuit.send(Message("ObjectDeselect", blocks, flags=PacketFlags.RELIABLE))
            ids_to_req = ids_to_req[255:]

        futures = []
        for local_id in local_ids:
            if local_id in unselected_ids:
                # Need to wait until we get our reply
                fut = self.state.register_future(local_id, ObjectUpdateType.PROPERTIES)
            else:
                # This was selected so we should already have up to date info
                fut = asyncio.Future()
                fut.set_result(self.lookup_localid(local_id))
            futures.append(fut)
        return futures

    def request_missing_objects(self) -> List[asyncio.Future[Object]]:
        return self.request_objects(self.state.missing_locals)

    def request_objects(self, local_ids: Union[int, Iterable[int]]) -> List[asyncio.Future[Object]]:
        """
        Request object local IDs, returning a list of awaitable handles for the objects

        Some may never be resolved, so use `asyncio.wait()` or `asyncio.wait_for()`.
        """
        if isinstance(local_ids, int):
            local_ids = (local_ids,)
        elif isinstance(local_ids, set):
            local_ids = tuple(local_ids)

        session = self._region.session()

        ids_to_req = local_ids
        while ids_to_req:
            self._region.circuit.send(Message(
                "RequestMultipleObjects",
                Block("AgentData", AgentID=session.agent_id, SessionID=session.id),
                *[Block("ObjectData", CacheMissType=0, ID=x) for x in ids_to_req[:255]],
                flags=PacketFlags.RELIABLE,
            ))
            ids_to_req = ids_to_req[255:]

        futures = []
        for local_id in local_ids:
            futures.append(self.state.register_future(local_id, ObjectUpdateType.OBJECT_UPDATE))
        return futures


class ObjectEvent:
    __slots__ = ("object", "updated", "update_type")

    object: Object
    updated: Set[str]
    update_type: ObjectUpdateType

    def __init__(self, obj: Object, updated: Set[str], update_type: ObjectUpdateType):
        self.object = obj
        self.updated = updated
        self.update_type = update_type

    @property
    def name(self) -> ObjectUpdateType:
        return self.update_type


class ClientWorldObjectManager:
    """Manages Objects for a session's whole world"""
    def __init__(self, session: BaseClientSession, settings: Settings, name_cache: Optional[NameCache]):
        self._session: BaseClientSession = session
        self._settings = settings
        self.name_cache = name_cache or NameCache()
        self.events: MessageHandler[ObjectEvent, ObjectUpdateType] = MessageHandler(take_by_default=False)
        self._fullid_lookup: Dict[UUID, Object] = {}
        self._avatars: Dict[UUID, Avatar] = {}
        self._avatar_objects: Dict[UUID, Object] = {}
        self._region_managers: Dict[int, ClientObjectManager] = {}
        message_handler = self._session.message_handler
        message_handler.subscribe("ObjectUpdate", self._handle_object_update)
        message_handler.subscribe("ImprovedTerseObjectUpdate",
                                  self._handle_terse_object_update)
        message_handler.subscribe("ObjectUpdateCompressed",
                                  self._handle_object_update_compressed)
        message_handler.subscribe("ObjectUpdateCached",
                                  self._handle_object_update_cached)
        message_handler.subscribe("CoarseLocationUpdate",
                                  self._handle_coarse_location_update)
        message_handler.subscribe("KillObject",
                                  self._handle_kill_object)
        message_handler.subscribe("ObjectProperties",
                                  self._handle_object_properties_generic)
        message_handler.subscribe("ObjectPropertiesFamily",
                                  self._handle_object_properties_generic)

    def lookup_fullid(self, full_id: UUID) -> Optional[Object]:
        return self._fullid_lookup.get(full_id, None)

    @property
    def all_objects(self) -> Iterable[Object]:
        return self._fullid_lookup.values()

    def lookup_avatar(self, full_id: UUID) -> Optional[Avatar]:
        return {a.FullID: a for a in self.all_avatars}.get(full_id, None)

    @property
    def all_avatars(self) -> Iterable[Avatar]:
        return tuple(self._avatars.values())

    def __len__(self):
        return len(self._fullid_lookup)

    def _get_region_state(self, handle: int) -> Optional[RegionObjectsState]:
        val = self._get_region_manager(handle)
        if val is None:
            return None
        return val.state

    def track_region_objects(self, handle: int):
        """Start tracking objects for a region"""
        if self._get_region_manager(handle) is None:
            self._region_managers[handle] = proxify(self._session.region_by_handle(handle).objects)

    def clear_region_objects(self, handle: int):
        """Handle signal that a region object manager was just cleared"""
        # Make sure they're gone from our lookup table
        for obj in tuple(self._fullid_lookup.values()):
            if obj.RegionHandle == handle:
                del self._fullid_lookup[obj.FullID]
        self._rebuild_avatar_objects()

    def _get_region_manager(self, handle: int) -> Optional[ClientObjectManager]:
        return self._region_managers.get(handle)

    def request_missing_objects(self) -> List[asyncio.Future[Object]]:
        futs = []
        for region in self._session.regions:
            futs.extend(region.objects.request_missing_objects())
        return futs

    def request_object_properties(self, objects: Union[Object, Sequence[Object]]) \
            -> List[asyncio.Future[Object]]:
        # Doesn't accept local ID unlike ObjectManager because they're ambiguous here.
        if isinstance(objects, Object):
            objects = (objects,)
        if not objects:
            return []

        # Has to be sent to the region they belong to, so split the objects out by region handle.
        objs_by_region = collections.defaultdict(list)
        for obj in objects:
            objs_by_region[obj.RegionHandle].append(obj)

        futs = []
        for region_handle, region_objs in objs_by_region.items():
            region_mgr = self._get_region_manager(region_handle)
            futs.extend(region_mgr.request_object_properties(region_objs))
        return futs

    async def load_ancestors(self, obj: Object, wait_time: float = 1.0):
        """
        Ensure that the entire chain of parents above this object is loaded

        Use this to make sure the object you're dealing with isn't orphaned and
        its RegionPosition can be determined.
        """
        region_mgr = self._get_region_manager(obj.RegionHandle)
        while obj.ParentID:
            if obj.Parent is None:
                await asyncio.wait_for(region_mgr.request_objects(obj.ParentID)[0], wait_time)
            obj = obj.Parent

    def clear(self):
        self._avatars.clear()
        for region_mgr in self._region_managers.values():
            region_mgr.clear()
        if self._fullid_lookup:
            LOG.warning(f"Had {len(self._fullid_lookup)} objects not tied to a region manager!")
        self._fullid_lookup.clear()
        self._rebuild_avatar_objects()
        self._region_managers.clear()

    def _update_existing_object(self, obj: Object, new_properties: dict, update_type: ObjectUpdateType,
                                msg: Optional[Message]):
        old_parent_id = obj.ParentID
        new_parent_id = new_properties.get("ParentID", obj.ParentID)
        old_local_id = obj.LocalID
        new_local_id = new_properties.get("LocalID", obj.LocalID)
        old_region_handle = obj.RegionHandle
        new_region_handle = new_properties.get("RegionHandle", obj.RegionHandle)
        old_region_state = self._get_region_state(old_region_handle)
        new_region_state = self._get_region_state(new_region_handle)

        actually_updated_props = set()

        if old_region_handle != new_region_handle:
            # The object just changed regions, we have to remove it from the old one.
            # Our LocalID will most likely change because, well, our locale changed.
            old_region_state.untrack_object(obj)
        elif old_local_id != new_local_id:
            # Our LocalID changed, and we deal with linkages to other prims by
            # LocalID association. Break any links since our LocalID is changing.
            # Could happen if we didn't mark an attachment prim dead and the parent agent
            # came back into the sim. Attachment FullIDs do not change across TPs,
            # LocalIDs do. This at least lets us partially recover from the bad state.
            new_localid = new_properties["LocalID"]
            LOG.warning(f"Got an update with new LocalID for {obj.FullID}, {obj.LocalID} != {new_localid}. "
                        f"May have mishandled a KillObject for a prim that left and re-entered region.")
            old_region_state.untrack_object(obj)
            obj.LocalID = new_localid
            old_region_state.track_object(obj)
            actually_updated_props |= {"LocalID"}

        actually_updated_props |= obj.update_properties(new_properties)

        if new_region_handle != old_region_handle:
            # Region just changed to this region, we should have untracked it before
            # so mark it tracked on this region. This should implicitly pick up any
            # orphans and handle parent ID changes.
            if new_region_state is not None:
                new_region_state.track_object(obj)
            else:
                # This will leave a regionless object in the global lookup dict, same as indra.
                LOG.warning(f"Tried to move object {obj!r} to unknown region {new_region_handle}")

            if obj.PCode == PCode.AVATAR:
                # `Avatar` instances are handled separately. Update all Avatar objects,
                # so we can deal with the RegionHandle change.
                self._rebuild_avatar_objects()
        elif new_parent_id != old_parent_id:
            # Parent ID changed, but we're in the same region
            new_region_state.handle_object_reparented(obj, old_parent_id=old_parent_id)

        if actually_updated_props and new_region_state is not None:
            self._run_object_update_hooks(obj, actually_updated_props, update_type, msg)

    def _track_new_object(self, region: RegionObjectsState, obj: Object, msg: Message):
        region.track_object(obj)
        self._fullid_lookup[obj.FullID] = obj
        if obj.PCode == PCode.AVATAR:
            self._avatar_objects[obj.FullID] = obj
            self._rebuild_avatar_objects()
        self._run_object_update_hooks(obj, set(obj.to_dict().keys()), ObjectUpdateType.OBJECT_UPDATE, msg)

    def _kill_object_by_local_id(self, region_state: RegionObjectsState, local_id: int):
        obj = region_state.lookup_localid(local_id)
        region_state.missing_locals -= {local_id}
        child_ids: Sequence[int]

        if obj:
            self._run_kill_object_hooks(obj)
            child_ids = obj.ChildIDs
        else:
            LOG.debug(f"Tried to kill unknown object {local_id}")
            # Kill any pending futures it might have had since untrack_object()
            # won't be called.
            region_state.cancel_futures(local_id)
            # If it had any orphans, they need to die.
            child_ids = region_state.collect_orphans(local_id)

        # KillObject implicitly kills descendents
        # This may mutate child_ids, use the reversed iterator so we don't
        # invalidate the iterator during removal.
        for child_id in reversed(child_ids):
            # indra special-cases avatar PCodes and doesn't mark them dead
            # due to cascading kill. Is this correct? Do avatars require
            # explicit kill? Does this imply ParentID = 0 or do we need
            # an explicit follow-up update?
            child_obj = region_state.lookup_localid(child_id)
            if child_obj and child_obj.PCode == PCode.AVATAR:
                continue
            self._kill_object_by_local_id(region_state, child_id)

        # Have to do this last, since untracking will clear child IDs
        if obj:
            region_state.untrack_object(obj)
            self._fullid_lookup.pop(obj.FullID, None)
            if obj.PCode == PCode.AVATAR:
                self._avatar_objects.pop(obj.FullID, None)
                self._rebuild_avatar_objects()

    def _handle_object_update(self, msg: Message):
        seen_locals = []
        handle = msg["RegionData"]["RegionHandle"]
        region_state = self._get_region_state(handle)
        for block in msg['ObjectData']:
            object_data = normalize_object_update(block, handle)
            seen_locals.append(object_data["LocalID"])
            if region_state is None:
                LOG.warning(f"Got ObjectUpdate for unknown region {handle}: {object_data!r}")
            # Do a lookup by FullID, if an object with this FullID already exists anywhere in
            # our view of the world then we want to move it to this region.
            obj = self.lookup_fullid(object_data["FullID"])
            if obj:
                self._update_existing_object(obj, object_data, ObjectUpdateType.OBJECT_UPDATE, msg)
            else:
                if region_state is None:
                    continue
                self._track_new_object(region_state, Object(**object_data), msg)
        msg.meta["ObjectUpdateIDs"] = tuple(seen_locals)

    def _handle_terse_object_update(self, msg: Message):
        seen_locals = []
        handle = msg["RegionData"]["RegionHandle"]
        region_state = self._get_region_state(handle)
        for block in msg['ObjectData']:
            object_data = normalize_terse_object_update(block, handle)

            if region_state is None:
                LOG.warning(f"Got ImprovedTerseObjectUpdate for unknown region {handle}: {object_data!r}")
                continue

            obj = region_state.lookup_localid(object_data["LocalID"])
            # Can only update existing object with this message
            if obj:
                # Need the Object as context because decoding state requires PCode.
                state_deserializer = ObjectStateSerializer.deserialize
                object_data["State"] = state_deserializer(ctx_obj=obj, val=object_data["State"])
                self._update_existing_object(obj, object_data, ObjectUpdateType.OBJECT_UPDATE, msg)
            else:
                if region_state:
                    region_state.missing_locals.add(object_data["LocalID"])
                LOG.debug(f"Received terse update for unknown object {object_data['LocalID']}")
            seen_locals.append(object_data["LocalID"])

        msg.meta["ObjectUpdateIDs"] = tuple(seen_locals)

    def _handle_object_update_cached(self, msg: Message):
        seen_locals = []
        missing_locals = set()
        handle = msg["RegionData"]["RegionHandle"]
        region_state = self._get_region_state(handle)
        for block in msg['ObjectData']:
            seen_locals.append(block["ID"])
            update_flags = block.deserialize_var("UpdateFlags", make_copy=False)

            if region_state is None:
                LOG.warning(f"Got ObjectUpdateCached for unknown region {handle}: {block!r}")
                continue

            # Check if we already know about the object
            obj = region_state.lookup_localid(block["ID"])
            if obj is not None and obj.CRC == block["CRC"]:
                self._update_existing_object(obj, {
                    "UpdateFlags": update_flags,
                    "RegionHandle": handle,
                }, ObjectUpdateType.OBJECT_UPDATE, msg)
                continue

            cached_obj_data = self._lookup_cache_entry(handle, block["ID"], block["CRC"])
            if cached_obj_data is not None:
                cached_obj = normalize_object_update_compressed_data(cached_obj_data)
                cached_obj["UpdateFlags"] = update_flags
                cached_obj["RegionHandle"] = handle
                self._track_new_object(region_state, Object(**cached_obj), msg)
                continue

            # Don't know about it and wasn't cached.
            missing_locals.add(block["ID"])
        if region_state:
            region_state.missing_locals.update(missing_locals)
        if missing_locals:
            self._handle_object_update_cached_misses(handle, missing_locals)
        msg.meta["ObjectUpdateIDs"] = tuple(seen_locals)

    def _handle_object_update_cached_misses(self, region_handle: int, missing_locals: Set[int]):
        """Handle an ObjectUpdateCached that referenced some un-cached local IDs"""
        region_mgr = self._get_region_manager(region_handle)
        region_mgr.request_objects(missing_locals)

    # noinspection PyUnusedLocal
    def _lookup_cache_entry(self, region_handle: int, local_id: int, crc: int) -> Optional[bytes]:
        return None

    def _handle_object_update_compressed(self, msg: Message):
        seen_locals = []
        handle = msg["RegionData"]["RegionHandle"]
        region_state = self._get_region_state(handle)
        for block in msg['ObjectData']:
            object_data = normalize_object_update_compressed(block, handle)
            seen_locals.append(object_data["LocalID"])
            if region_state is None:
                LOG.warning(f"Got ObjectUpdateCompressed for unknown region {handle}: {object_data!r}")
            obj = self.lookup_fullid(object_data["FullID"])
            if obj:
                self._update_existing_object(obj, object_data, ObjectUpdateType.OBJECT_UPDATE, msg)
            else:
                if region_state is None:
                    continue
                self._track_new_object(region_state, Object(**object_data), msg)
        msg.meta["ObjectUpdateIDs"] = tuple(seen_locals)

    def _handle_object_properties_generic(self, packet: Message):
        seen_locals = []
        for block in packet["ObjectData"]:
            object_properties = dict(block.items())
            if packet.name == "ObjectProperties":
                object_properties["TextureID"] = block.deserialize_var("TextureID")

            obj = self.lookup_fullid(block["ObjectID"])
            if obj:
                seen_locals.append(obj.LocalID)
                self._update_existing_object(obj, object_properties, ObjectUpdateType.PROPERTIES, packet)
            else:
                LOG.debug(f"Received {packet.name} for unknown {block['ObjectID']}")
        packet.meta["ObjectUpdateIDs"] = tuple(seen_locals)

    def _handle_kill_object(self, message: Message):
        seen_locals = []

        # Have to look up region based on sender, handle not sent in this message
        region = self._session.region_by_circuit_addr(message.sender)
        region_state = region.objects.state
        for block in message["ObjectData"]:
            self._kill_object_by_local_id(region_state, block["ID"])
            seen_locals.append(block["ID"])
        message.meta["ObjectUpdateIDs"] = tuple(seen_locals)

    def _handle_coarse_location_update(self, message: Message):
        # Have to look up region based on sender, handle not sent in this message
        region = self._session.region_by_circuit_addr(message.sender)
        region_state = region.objects.state
        region_state.coarse_locations.clear()

        coarse_locations: Dict[UUID, Vector3] = {}
        for agent_block, location_block in zip(message["AgentData"], message["Location"]):
            x, y, z = location_block["X"], location_block["Y"], location_block["Z"]
            coarse_locations[agent_block["AgentID"]] = Vector3(
                X=x,
                Y=y,
                # The z-axis is multiplied by 4 to obtain true Z location
                # The z-axis is also limited to 1020m in height
                # If z == 255 then the true Z is unknown.
                # http://wiki.secondlife.com/wiki/CoarseLocationUpdate
                Z=z * 4 if z != 255 else math.inf,
            )

        region_state.coarse_locations.update(coarse_locations)
        self._rebuild_avatar_objects()

    def _process_get_object_cost_response(self, parsed: dict):
        if "error" in parsed:
            return
        for object_id, object_costs in parsed.items():
            obj = self.lookup_fullid(UUID(object_id))
            if not obj:
                LOG.debug(f"Received ObjectCost for unknown {object_id}")
                continue
            obj.ObjectCosts.update(object_costs)
            self._run_object_update_hooks(obj, {"ObjectCosts"}, ObjectUpdateType.COSTS, None)

    def _run_object_update_hooks(self, obj: Object, updated_props: Set[str], update_type: ObjectUpdateType,
                                 msg: Optional[Message]):
        region_state = self._get_region_state(obj.RegionHandle)
        if region_state:
            region_state.resolve_futures(obj, update_type)
        else:
            LOG.warning(f"{obj} not tied to a region state")

        if obj.PCode == PCode.AVATAR and "NameValue" in updated_props:
            if obj.NameValue:
                self.name_cache.update(obj.FullID, obj.NameValue.to_dict())
        self.events.handle(ObjectEvent(obj, updated_props, update_type))

    def _run_kill_object_hooks(self, obj: Object):
        self.events.handle(ObjectEvent(obj, set(), ObjectUpdateType.KILL))

    def _rebuild_avatar_objects(self):
        # Get all avatars known through coarse locations and which region the location was in
        coarse_locations: Dict[UUID, Tuple[int, Vector3]] = {}
        for region_handle, region in self._region_managers.items():
            for av_key, location in region.state.coarse_locations.items():
                coarse_locations[av_key] = (region_handle, location)

        # Merge together avatars known through coarse locations or objects, with details for both
        current_av_details: Dict[UUID, Tuple[Optional[Tuple[int, Vector3]], Optional[Object]]] = {}
        for av_key in set(coarse_locations.keys()) | set(self._avatar_objects.keys()):
            details = (coarse_locations.get(av_key), self._avatar_objects.get(av_key))
            current_av_details[av_key] = details

        # Look for changes in avatars we're already tracking
        for existing_key in tuple(self._avatars.keys()):
            av = self._avatars[existing_key]
            if existing_key in current_av_details:
                # This avatar this exists, update it.
                coarse_pair, av_obj = current_av_details[existing_key]
                av.Object = av_obj
                if coarse_pair:
                    coarse_handle, coarse_location = coarse_pair
                    av.CoarseLocation = coarse_location
                    av.RegionHandle = coarse_handle
                    # If we have a real value for Z then throw away any stale guesses
                    if av.CoarseLocation.Z != math.inf:
                        av.GuessedZ = None
                if av_obj:
                    av.Object = av_obj
                    av.RegionHandle = av_obj.RegionHandle
            else:
                # Avatar isn't in coarse locations or objects, it's gone.
                self._avatars.pop(existing_key, None)
                av.Object = None
                av.CoarseLocation = None
                av.Valid = False

        # Check for any new avatars
        for av_key, (coarse_pair, av_obj) in current_av_details.items():
            if av_key in self._avatars:
                # Already handled in the update step above
                continue
            region_handle = None
            coarse_location = None
            if coarse_pair:
                region_handle, coarse_location = coarse_pair
            if av_obj:
                # Prefer the region handle from the Object if we have one
                region_handle = av_obj.RegionHandle
            assert region_handle is not None
            self._avatars[av_key] = Avatar(
                full_id=av_key,
                region_handle=region_handle,
                resolved_name=self.name_cache.lookup(av_key, create_if_none=True),
                coarse_location=coarse_location,
                obj=av_obj,
            )


class RegionObjectsState:
    """
    Internal class for tracking Object state within a specific region

    Should only be directly used by the world and region ObjectManagers.
    """

    __slots__ = (
        "handle", "missing_locals", "_orphans", "localid_lookup", "coarse_locations",
        "_object_futures"
    )

    def __init__(self):
        self.missing_locals = set()
        self.localid_lookup: Dict[int, Object] = {}
        self.coarse_locations: Dict[UUID, Vector3] = {}
        self._object_futures: Dict[Tuple[int, int], List[asyncio.Future]] = {}
        self._orphans: Dict[int, List[int]] = collections.defaultdict(list)

    def clear(self):
        """Called by the owning ObjectManager when it knows the region is going away"""
        for fut in tuple(itertools.chain(*self._object_futures.values())):
            fut.cancel()
        self._object_futures.clear()
        self._orphans.clear()
        self.coarse_locations.clear()
        self.missing_locals.clear()
        self.localid_lookup.clear()

    def lookup_localid(self, localid: int) -> Optional[Object]:
        return self.localid_lookup.get(localid)

    def track_object(self, obj: Object):
        """Assign ownership of Object to this region"""
        obj_same_localid = self.localid_lookup.get(obj.LocalID)
        if obj_same_localid:
            LOG.error(f"Clobbering existing object with LocalID {obj.LocalID}! "
                      f"{obj.to_dict()} clobbered {obj_same_localid.to_dict()}")
        self.localid_lookup[obj.LocalID] = obj
        # If it was missing, it's not missing anymore.
        self.missing_locals -= {obj.LocalID}

        self._parent_object(obj)

        # Adopt any of our orphaned child objects.
        for orphan_local in self.collect_orphans(obj.LocalID):
            child_obj = self.localid_lookup.get(orphan_local)
            # Shouldn't be any dead children in the orphanage
            assert child_obj is not None
            self._parent_object(child_obj)

    def untrack_object(self, obj: Object):
        """
        Take ownership of an Object from this region

        Can happen due to the object being killed, or due to it moving to another region
        """
        former_child_ids = obj.ChildIDs[:]
        for child_id in former_child_ids:
            child_obj = self.localid_lookup.get(child_id)
            assert child_obj is not None
            self._unparent_object(child_obj, child_obj.ParentID)

        # Place any remaining unkilled children in the orphanage
        for child_id in former_child_ids:
            self._track_orphan(child_id, obj.LocalID)

        assert not obj.ChildIDs

        # Make sure the parent knows we went away
        self._unparent_object(obj, obj.ParentID)
        # Object doesn't belong to this region anymore and won't receive
        # any updates, cancel any pending futures
        self.cancel_futures(obj.LocalID)

        del self.localid_lookup[obj.LocalID]

    def _parent_object(self, obj: Object, insert_at_head=False):
        """Create any links to ancestor Objects for obj"""
        if obj.ParentID:
            parent = self.localid_lookup.get(obj.ParentID)
            if parent is not None:
                assert obj.LocalID not in parent.ChildIDs
                # Link order is never explicitly passed to clients, so we have to do
                # some nasty guesswork based on order of received initial ObjectUpdates
                # Note that this is broken in the viewer as well, and there doesn't seem
                # to be a foolproof way to get this.
                idx = 0 if insert_at_head else len(parent.ChildIDs)
                parent.ChildIDs.insert(idx, obj.LocalID)
                parent.Children.insert(idx, obj)
                obj.Parent = weakref.proxy(parent)
            else:
                # We have a parent, but we don't have an Object for it yet
                self.missing_locals.add(obj.ParentID)
                self._track_orphan(obj.LocalID, parent_id=obj.ParentID)
                obj.Parent = None
                LOG.debug(f"{obj.LocalID} updated with parent {obj.ParentID}, but parent wasn't found!")

    def _unparent_object(self, obj: Object, old_parent_id: int):
        """Break any links to ancestor Objects for obj"""
        obj.Parent = None
        if old_parent_id:
            # Had a parent, remove this from the child and orphan lists.
            removed = self._untrack_orphan(obj, old_parent_id)

            old_parent = self.localid_lookup.get(old_parent_id)
            if old_parent:
                if obj.LocalID in old_parent.ChildIDs:
                    idx = old_parent.ChildIDs.index(obj.LocalID)
                    del old_parent.ChildIDs[idx]
                    del old_parent.Children[idx]
                else:
                    # Something is very broken if this happens
                    LOG.warning(f"Changing parent of {obj.LocalID}, but old parent didn't correctly adopt, "
                                f"was {'' if removed else 'not '}in orphan list")
            else:
                LOG.debug(f"Changing parent of {obj.LocalID}, but couldn't find old parent")

    def handle_object_reparented(self, obj: Object, old_parent_id: int):
        """Recreate any links to ancestor Objects for obj due to parent changes"""
        self._unparent_object(obj, old_parent_id)
        self._parent_object(obj, insert_at_head=True)

    def collect_orphans(self, parent_localid: int) -> Sequence[int]:
        """Take ownership of any orphan IDs belonging to parent_localid"""
        return self._orphans.pop(parent_localid, [])

    def _track_orphan(self, local_id: int, parent_id: int):
        if len(self._orphans) > 100:
            LOG.warning(f"Orphaned object dict is getting large: {len(self._orphans)}")
        self._orphans[parent_id].append(local_id)

    def _untrack_orphan(self, obj: Object, parent_id: int):
        """Remove obj from parent_id's list of orphans if present"""
        if parent_id not in self._orphans:
            return False
        orphan_list = self._orphans[parent_id]
        removed = False
        if obj.LocalID in orphan_list:
            orphan_list.remove(obj.LocalID)
            removed = True
        # List is empty now, get rid of it.
        if not orphan_list:
            del self._orphans[parent_id]
        return removed

    def register_future(self, local_id: int, future_type: ObjectUpdateType) -> asyncio.Future[Object]:
        fut = asyncio.Future()
        fut_key = (local_id, future_type)
        local_futs = self._object_futures.get(fut_key, [])
        local_futs.append(fut)
        self._object_futures[fut_key] = local_futs
        fut.add_done_callback(local_futs.remove)
        return fut

    def resolve_futures(self, obj: Object, update_type: ObjectUpdateType):
        futures = self._object_futures.get((obj.LocalID, update_type), [])
        for fut in futures[:]:
            fut.set_result(obj)

    def cancel_futures(self, local_id: int):
        # Object went away, so need to kill any pending futures.
        for fut_key, futs in self._object_futures.items():
            if fut_key[0] == local_id:
                for fut in futs:
                    fut.cancel()
                break


class LocationType(enum.IntEnum):
    NONE = enum.auto()
    COARSE = enum.auto()
    EXACT = enum.auto()


class Avatar:
    """Wrapper for an avatar known through ObjectUpdate or CoarseLocationUpdate"""

    def __init__(
            self,
            full_id: UUID,
            region_handle: int,
            obj: Optional["Object"] = None,
            coarse_location: Optional[Vector3] = None,
            resolved_name: Optional[NameCacheEntry] = None,
    ):
        self.FullID: UUID = full_id
        self.Object: Optional["Object"] = obj
        self.RegionHandle: int = region_handle
        # TODO: Allow hooking into getZOffsets FS bridge response
        #  to fill in the Z axis if it's infinite
        self.CoarseLocation = coarse_location
        self.Valid = True
        self.GuessedZ: Optional[float] = None
        self._resolved_name = resolved_name

    @property
    def LocationType(self) -> "LocationType":
        if self.Object and self.Object.AncestorsKnown:
            return LocationType.EXACT
        if self.CoarseLocation is not None:
            return LocationType.COARSE
        return LocationType.NONE

    @property
    def RegionPosition(self) -> Vector3:
        if self.Object and self.Object.AncestorsKnown:
            return self.Object.RegionPosition
        if self.CoarseLocation is not None:
            if self.CoarseLocation.Z == math.inf and self.GuessedZ is not None:
                coarse = self.CoarseLocation
                return Vector3(coarse.X, coarse.Y, self.GuessedZ)
            return self.CoarseLocation
        raise ValueError(f"Avatar {self.FullID} has no known position")

    @property
    def GlobalPosition(self) -> Vector3:
        return self.RegionPosition + handle_to_global_pos(self.RegionHandle)

    @property
    def Name(self) -> Optional[str]:
        if not self._resolved_name:
            return None
        return str(self._resolved_name)

    @property
    def PreferredName(self) -> Optional[str]:
        if not self._resolved_name:
            return None
        return self._resolved_name.preferred_name

    @property
    def DisplayName(self) -> Optional[str]:
        if not self._resolved_name:
            return None
        return self._resolved_name.display_name

    @property
    def LegacyName(self) -> Optional[str]:
        if not self._resolved_name:
            return None
        return self._resolved_name.legacy_name

    def __repr__(self):
        loc_str = str(self.RegionPosition) if self.LocationType != LocationType.NONE else "?"
        return f"<{self.__class__.__name__} {self.FullID} {self.Name!r} @ {loc_str}>"



import abc
import ast
import typing

from arpeggio import Optional, ZeroOrMore, EOF, \
    ParserPython, PTNodeVisitor, visit_parse_tree, RegExMatch, OneOrMore


def literal():
    return [
        # Nightmare. str or bytes literal.
        # https://stackoverflow.com/questions/14366401/#comment79795017_14366904
        RegExMatch(r'''b?(\"\"\"|\'\'\'|\"|\')((?<!\\)(\\\\)*\\\1|.)*?\1'''),
        # base16
        RegExMatch(r'0x[0-9a-fA-F]+'),
        # base10 int or float.
        RegExMatch(r'\d+(\.\d+)?'),
        "None",
        "True",
        "False",
        # vector3 (tuple)
        RegExMatch(r'\(\s*\d+(\.\d+)?\s*,\s*\d+(\.\d+)?\s*,\s*\d+(\.\d+)?\s*\)'),
        # vector4 (tuple)
        RegExMatch(r'\(\s*\d+(\.\d+)?\s*,\s*\d+(\.\d+)?\s*,\s*\d+(\.\d+)?\s*,\s*\d+(\.\d+)?\s*\)'),
    ]


def identifier():
    # Identifiers are allowed to have "-". It's not a special character
    # in our grammar, and we expect them to show up some places, like header names.
    return RegExMatch(r'[a-zA-Z*]([a-zA-Z0-9_*-]+)?')


def field_specifier():
    return identifier, ZeroOrMore('.', identifier)


def unary_field_specifier():
    return field_specifier


def unary_expression():
    return Optional(["!"]), [unary_field_specifier, ("(", expression, ")")]


def meta_field_specifier():
    return "Meta", OneOrMore(".", identifier)


def enum_field_specifier():
    return identifier, ".", identifier


def compare_val():
    return [literal, meta_field_specifier, enum_field_specifier]


def binary_expression():
    return field_specifier, ["==", "!=", "^=", "$=", "~=", ">", ">=", "<", "<=", "&"], compare_val


def term():
    return [binary_expression, unary_expression]


def expression():
    return term, ZeroOrMore(["||", "&&"], expression)


def message_filter():
    return expression, EOF


class MatchResult(typing.NamedTuple):
    result: bool
    fields: typing.List[typing.Tuple]

    def __bool__(self):
        return self.result


class BaseFilterNode(abc.ABC):
    @abc.abstractmethod
    def match(self, msg, short_circuit=True) -> MatchResult:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def children(self):
        raise NotImplementedError()


class UnaryFilterNode(BaseFilterNode, abc.ABC):
    def __init__(self, node):
        self.node: BaseFilterNode = node

    @property
    def children(self):
        return self.node


class BinaryFilterNode(BaseFilterNode, abc.ABC):
    def __init__(self, left_node, right_node):
        super().__init__()
        self.left_node: BaseFilterNode = left_node
        self.right_node: BaseFilterNode = right_node

    @property
    def children(self):
        return self.left_node, self.right_node


class UnaryNotFilterNode(UnaryFilterNode):
    def match(self, msg, short_circuit=True) -> MatchResult:
        # Should we pass fields up here? Maybe not.
        return MatchResult(not self.node.match(msg, short_circuit), [])


class OrFilterNode(BinaryFilterNode):
    def match(self, msg, short_circuit=True) -> MatchResult:
        left_match = self.left_node.match(msg, short_circuit)
        if left_match and short_circuit:
            return MatchResult(True, left_match.fields)

        right_match = self.right_node.match(msg, short_circuit)
        if right_match and short_circuit:
            return MatchResult(True, right_match.fields)

        if left_match or right_match:
            # Fine since fields should be empty when result=False
            return MatchResult(True, left_match.fields + right_match.fields)
        return MatchResult(False, [])


class AndFilterNode(BinaryFilterNode):
    def match(self, msg, short_circuit=True) -> MatchResult:
        left_match = self.left_node.match(msg, short_circuit)
        if not left_match:
            return MatchResult(False, [])
        right_match = self.right_node.match(msg, short_circuit)
        if not right_match:
            return MatchResult(False, [])
        return MatchResult(True, left_match.fields + right_match.fields)


class MessageFilterNode(BaseFilterNode):
    def __init__(self, selector: typing.Sequence[str], operator: typing.Optional[str], value: typing.Optional):
        self.selector = selector
        self.operator = operator
        self.value = value

    def match(self, msg, short_circuit=True) -> MatchResult:
        return msg.matches(self, short_circuit)

    @property
    def children(self):
        return self.selector, self.operator, self.value


class MetaFieldSpecifier(tuple):
    pass


class EnumFieldSpecifier(typing.NamedTuple):
    enum_name: str
    field_name: str


class LiteralValue:
    """Only exists because we can't return `None` in a visitor, need to box it"""
    def __init__(self, value):
        self.value = value


class MessageFilterVisitor(PTNodeVisitor):
    def visit_identifier(self, node, _children):
        return str(node.value)

    def visit_field_specifier(self, _node, children):
        return children

    def visit_literal(self, node, _children):
        return LiteralValue(ast.literal_eval(node.value))

    def visit_meta_field_specifier(self, _node, children):
        return MetaFieldSpecifier(children)

    def visit_enum_field_specifier(self, _node, children):
        return EnumFieldSpecifier(*children)

    def visit_unary_field_specifier(self, _node, children):
        # Looks like a bare field specifier with no operator
        return MessageFilterNode(tuple(children), None, None)

    def visit_unary_expression(self, _node, children):
        if len(children) == 1:
            if isinstance(children[0], BaseFilterNode):
                return children[0]
            else:
                raise ValueError("What?")
        # Might have a unary !
        if children[0] == "!":
            return UnaryNotFilterNode(children[1])
        else:
            raise ValueError(f"Unrecognized unary prefix {children[0]}")

    def visit_binary_expression(self, _node, children):
        return MessageFilterNode(tuple(children[0]), children[1], children[2])

    def visit_expression(self, _node, children):
        if self.debug:
            print("Expression {}".format(children))
        if len(children) > 1:
            if children[1] == "&&":
                return AndFilterNode(children[0], children[2])
            elif children[1] == "||":
                return OrFilterNode(children[0], children[2])
            else:
                raise ValueError(f"Unrecognized operator {children[1]}")
        return children[0]


def compile_filter(filter_str) -> BaseFilterNode:
    filter_str = filter_str.strip()
    if not filter_str.strip():
        filter_str = "*"
    elif filter_str == "!":
        filter_str = "!*"
    parser = ParserPython(message_filter)
    parse_tree = parser.parse(filter_str)
    return visit_parse_tree(parse_tree, MessageFilterVisitor())
